LABEL_RANGE = {
    "car": 50,
    "truck": 50,
    "bus": 50,
    "trailer": 50,
    "construction_vehicle": 50,
    "pedestrian": 40,
    "motorcycle": 40,
    "bicycle": 40,
    "traffic_cone": 30,
    "barrier": 30,
}
DIST_FCN = "center_distance"
DIST_THS = [0.5, 1.0, 2.0, 4.0]
MIN_RECALL = 0.1
MIN_PRECISION = 0.1
MAX_BOXES_PER_SAMPLE = 500
MEAN_AP_WEIGHT = 5  # constant used in NuScenes Detection Score (NDS) to scale
# mAP



import logging
import operator
from typing import Dict, List, Tuple

import numpy as np
from nuscenes.eval.detection.data_classes import DetectionMetrics

from datasetinsights.io.bbox import BBox3d

from .average_precision_config import (
    DIST_FCN,
    DIST_THS,
    LABEL_RANGE,
    MAX_BOXES_PER_SAMPLE,
    MEAN_AP_WEIGHT,
    MIN_PRECISION,
    MIN_RECALL,
)
from .base import EvaluationMetric

logger = logging.getLogger(__name__)


class DetectionConfig:
    """ Data class that specifies the detection evaluation settings. """

    def __init__(
        self,
        class_range: Dict[str, int] = LABEL_RANGE,
        dist_fcn: str = DIST_FCN,
        dist_ths: List[float] = DIST_THS,
        min_recall: float = MIN_RECALL,
        min_precision: float = MIN_PRECISION,
        max_boxes_per_sample: float = MAX_BOXES_PER_SAMPLE,
        mean_ap_weight: int = MEAN_AP_WEIGHT,
    ):
        self.class_range = class_range
        self.dist_fcn = dist_fcn
        self.dist_ths = dist_ths
        self.min_recall = min_recall
        self.min_precision = min_precision
        self.max_boxes_per_sample = max_boxes_per_sample
        self.mean_ap_weight = mean_ap_weight

        self.labels = self.class_range.keys()

    def __eq__(self, other):
        eq = True
        for key in self.serialize().keys():
            eq = eq and np.array_equal(getattr(self, key), getattr(other, key))
        return eq

    def serialize(self) -> dict:
        """ Serialize instance into json-friendly format. """
        return {
            "class_range": self.class_range,
            "dist_fcn": self.dist_fcn,
            "dist_ths": self.dist_ths,
            "min_recall": self.min_recall,
            "min_precision": self.min_precision,
            "max_boxes_per_sample": self.max_boxes_per_sample,
            "mean_ap_weight": self.mean_ap_weight,
        }

    @classmethod
    def deserialize(cls, content):
        """ Initialize from serialized dictionary. """
        return cls(
            content["class_range"],
            content["dist_fcn"],
            content["dist_ths"],
            content["min_recall"],
            content["min_precision"],
            content["max_boxes_per_sample"],
            content["mean_ap_weight"],
        )


def calc_ap(*, precision, min_recall: float, min_precision: float) -> float:
    """ Calculated average precision. """

    # assert 0 <= min_precision < 1
    # assert 0 <= min_recall <= 1

    prec = np.copy(precision)
    prec = prec[
        round(100 * min_recall) + 1 :
    ]  # Clip low recalls. +1 to exclude the min
    # recall bin.
    prec -= min_precision  # Clip low precision
    prec[prec < 0] = 0
    return float(np.mean(prec)) / (1.0 - min_precision)


def center_distance(*, gt_box: BBox3d, pred_box: BBox3d) -> float:
    """
    L2 distance between the box centers (xy only).
    :param gt_box: GT annotation sample.
    :param pred_box: Predicted sample.
    :return: L2 distance.
    """
    return np.linalg.norm(
        np.array(pred_box.translation[:2]) - np.array(gt_box.translation[:2])
    )


def _count_class_examples(
    boxes: Dict[str, Tuple[List[BBox3d], List[BBox3d]]], label
):
    npos = 0
    for image, (predicted_bbs, gt_bbs) in boxes.items():
        for gt_bounding_box in gt_bbs:
            if gt_bounding_box.label == label:
                npos += 1
    return npos


class AveragePrecision(EvaluationMetric):
    def __init__(self, config: DetectionConfig = None):
        self._num_examples = 0
        self.cfg = config or DetectionConfig()
        self.boxes = {}

    # todo need update method which accumulates serialized boxes
    def update(self, boxes):
        self.boxes.update(boxes)

    def reset(self):
        self.boxes = {}

    def _calculate_label_ap_simple(
        self,
        boxes: Dict[str, Tuple[List[BBox3d], List[BBox3d]]],
        label,
        dist_th,
    ):
        """
            calculate the average precision for a single class at a single
            distance threshold.
            :param boxes: keys -> (list of predicted boxes, list of ground
            truth boxes)
            :param label: name of the class used to calculate ap
            :param dist_th: distance threshold. Distance from center of the
            predicted box using birds-eye-view to
            ground truth box's center using birds eye view
            :return: average precesion for a single class at a single distance
             threshold
            """
        tp = []  # Accumulator of true positives
        fp = []  # Accumulator of false positives
        npos = _count_class_examples(boxes=boxes, label=label)
        for img, (predicted_bbs, gt_bbs) in boxes.items():
            predicted_bbs.sort(key=operator.attrgetter("score"), reverse=True)
            matches = set()
            for predicted_bb_index, predicted_box in enumerate(predicted_bbs):
                min_dist = np.inf
                for gt_idx, gt_box in enumerate(gt_bbs):
                    if gt_box.label == label and gt_idx not in matches:
                        this_distance = center_distance(
                            gt_box=gt_box, pred_box=predicted_box
                        )
                        if this_distance < min_dist:
                            min_dist = this_distance
                    is_match = min_dist < dist_th
                    if is_match:
                        tp.append(1)
                        fp.append(0)
                    else:
                        tp.append(0)
                        fp.append(1)
        if np.max(tp) == 0:
            return 0
        else:
            tp = np.cumsum(tp).astype(np.float)
            fp = np.cumsum(fp).astype(np.float)
            prec = tp / (fp + tp)
            rec = tp / float(npos)
            rec_interp = np.linspace(
                0, 1, 101
            )  # 101 steps, from 0% to 100% recall.
            prec = np.interp(rec_interp, rec, prec, right=0)
        ap = calc_ap(
            precision=prec,
            min_recall=self.cfg.min_recall,
            min_precision=self.cfg.min_precision,
        )
        return ap

    def compute(
        self, boxes: Dict[str, Tuple[List[BBox3d], List[BBox3d]]] = None
    ) -> Dict[str, float]:
        """
        calculate the mean ap for all classes over all distance thresholds
        defined in the config. The equation is
        described in the second figure in the nuscenes paper
        https://arxiv.org/pdf/1903.11027.pdf It is the normalized
        sum of the ROC curves for each class and distance.
        :param boxes: the predicted and ground truth bounding boxes per sample.
        :return: dictionary mapping each label to it's average precision
        (averaged across all distance thresholds for
        that label)
        """
        if boxes is None:
            boxes = self.boxes
        metrics = DetectionMetrics(self.cfg)
        for label in self.cfg.labels:
            for dist_th in self.cfg.dist_ths:
                ap = self._calculate_label_ap_simple(
                    boxes=boxes, dist_th=dist_th, label=label
                )
                metrics.add_label_ap(label, dist_th, ap)
        return metrics.mean_dist_aps



from abc import ABCMeta, abstractmethod

from tensorboardX import SummaryWriter

import datasetinsights.constants as const
from datasetinsights.io.checkpoint import EstimatorCheckpoint
from datasetinsights.io.kfp_output import KubeflowPipelineWriter


def create_estimator(
    name,
    config,
    *,
    tb_log_dir=None,
    no_cuda=None,
    checkpoint_dir=None,
    kfp_metrics_dir=const.DEFAULT_KFP_METRICS_DIR,
    kfp_metrics_filename=const.DEFAULT_KFP_METRICS_FILENAME,
    no_val=None,
    **kwargs,
):
    """Create a new instance of the estimators subclass

    Args:
        name (str): unique identifier for a estimators subclass
        config (dict): parameters specific to each estimators subclass
            used to create a estimators instance

    Returns:
        an instance of the specified estimators subclass
    """

    estimators_cls = _find_estimator(name)

    # todo this makes it so that we lose the tensorboard
    #  writer of non-master processes which could make debugging harder

    writer = SummaryWriter(tb_log_dir)
    kfp_writer = KubeflowPipelineWriter(
        filename=kfp_metrics_filename, filepath=kfp_metrics_dir,
    )
    checkpointer = EstimatorCheckpoint(
        estimator_name=name, checkpoint_dir=checkpoint_dir, distributed=False,
    )

    return estimators_cls(
        config=config,
        writer=writer,
        kfp_writer=kfp_writer,
        checkpointer=checkpointer,
        logdir=tb_log_dir,
        no_cuda=no_cuda,
        no_val=no_val,
        kfp_metrics_dir=kfp_metrics_dir,
        kfp_metrics_filename=kfp_metrics_filename,
        **kwargs,
    )


def _find_estimator(name):
    """Find Estimator subclass based on the given name

    Args:
        name (str): unique identifier for a estimators subclass

    Returns:
        a label of the specified estimators subclass
    """
    estimators_classes = Estimator.__subclasses__()
    estimators_names = [e.__name__ for e in estimators_classes]
    if name in estimators_names:
        estimators_cls = estimators_classes[estimators_names.index(name)]
        return estimators_cls
    else:
        raise NotImplementedError(f"Unknown Estimator class {name}!")


class Estimator(metaclass=ABCMeta):
    """Abstract base class for estimator.

    An estimator is the master class of all modeling operations. At minimum,
    it includes:

    1. input data and output data transformations (e.g. input image cropping,
    remove unused output labels...) when applicable.
    2. neural network graph (model) for either pytorch or tensorflow.
    3. procedures to execute model training and evaluation.

    One estimator could support multiple tasks (e.g. Mask R-CNN can be used for
    semantic segmentation and object detection)
    """

    @abstractmethod
    def train(self, **kwargs):
        """Abstract method to train estimators
        """
        raise NotImplementedError("Subclass needs to implement this method")

    @abstractmethod
    def evaluate(self, **kwargs):
        """Abstract method to evaluate estimators
        """
        raise NotImplementedError("Subclass needs to implement this method")



""" Simulation Dataset Catalog
"""
import glob
import logging
import os
import shutil
from pathlib import Path

from PIL import Image
from sklearn.model_selection import train_test_split

from datasetinsights.datasets.unity_perception import (
    AnnotationDefinitions,
    Captures,
)
from datasetinsights.datasets.unity_perception.tables import SCHEMA_VERSION
from datasetinsights.io.bbox import BBox2D

from .base import Dataset
from .exceptions import DatasetNotFoundError

logger = logging.getLogger(__name__)

DEFAULT_TRAIN_SPLIT_RATIO = 0.9
TRAIN = "train"
VAL = "val"
ALL = "all"
VALID_SPLITS = (TRAIN, VAL, ALL)


def _get_split(*, split, catalog, train_percentage=0.9, random_seed=47):
    """

    Args:
        split (str): can be 'train', 'val' or 'all'
        catalog (pandas Dataframe): dataframe which will be divided into splits
        train_percentage (float): percentage of dataframe to put in train split
        random_seed (int): random seed used for splitting dataset into train
        and val

    Returns: catalog (dataframe) divided into correct split

    """
    if split == ALL:
        logger.info(f"split specified was 'all' using entire synthetic dataset")
        return catalog
    train, val = train_test_split(
        catalog, train_size=train_percentage, random_state=random_seed
    )
    if split == TRAIN:
        catalog = train
        catalog.index = [i for i in range(len(catalog))]
        logger.info(
            f"split specified was {TRAIN}, using "
            f"{train_percentage*100:.2f}% of the dataset"
        )
        return catalog
    elif split == VAL:
        catalog = val
        catalog.index = [i for i in range(len(catalog))]
        logger.info(
            f"split specified was {VAL} using "
            f"{(1-train_percentage)*100:.2f}% of the dataset "
        )
        return catalog
    else:
        raise ValueError(
            f"split provided was {split} but only valid "
            f"splits are: {VALID_SPLITS}"
        )


def read_bounding_box_2d(annotation, label_mappings=None):
    """Convert dictionary representations of 2d bounding boxes into objects
    of the BBox2D class

    Args:
        annotation (List[dict]): 2D bounding box annotation
        label_mappings (dict): a dict of {label_id: label_name} mapping

    Returns:
        A list of 2D bounding box objects
    """
    bboxes = []
    for b in annotation:
        label_id = b["label_id"]
        x = b["x"]
        y = b["y"]
        w = b["width"]
        h = b["height"]
        if label_mappings and label_id not in label_mappings:
            continue
        box = BBox2D(label=label_id, x=x, y=y, w=w, h=h)
        bboxes.append(box)

    return bboxes


class SynDetection2D(Dataset):
    """Synthetic dataset for 2D object detection.

    During the class instantiation, it would check whether the data files
    such as annotations.json, images.png are present, if not it'll check
    whether a compressed dataset file is present which contains the necessary
    files, if not it'll raise an error.

    See synthetic dataset schema documentation for more details.
    <https://datasetinsights.readthedocs.io/en/latest/Synthetic_Dataset_Schema.html>

    Attributes:
        catalog (list): catalog of all captures in this dataset
        transforms: callable transformation that applies to a pair of
            capture, annotation. Capture is the information captured by the
            sensor, in this case an image, and annotations, which in this
            dataset are 2d bounding box coordinates and labels.
        split (str): indicate split type of the dataset (train|val|test)
        label_mappings (dict): a dict of {label_id: label_name} mapping
    """

    ARCHIVE_FILE = "SynthDet.zip"
    SUBFOLDER = "synthetic"

    def __init__(
        self,
        *,
        data_path=None,
        split="all",
        transforms=None,
        version=SCHEMA_VERSION,
        def_id=4,
        train_split_ratio=DEFAULT_TRAIN_SPLIT_RATIO,
        random_seed=47,
        **kwargs,
    ):
        """
        Args:
            data_path (str): Directory of the dataset
            transforms: callable transformation that applies to a pair of
            capture, annotation.
            version(str): synthetic dataset schema version
            def_id (int): annotation definition id used to filter results
            random_seed (int): random seed used for splitting dataset into
                train and val
        """
        self._data_path = self._preprocess_dataset(data_path)

        captures = Captures(self._data_path, version)
        annotation_definition = AnnotationDefinitions(self._data_path, version)
        catalog = captures.filter(def_id)
        self.catalog = self._cleanup(catalog)
        init_definition = annotation_definition.get_definition(def_id)
        self.label_mappings = {
            m["label_id"]: m["label_name"] for m in init_definition["spec"]
        }

        if split not in VALID_SPLITS:
            raise ValueError(
                f"split provided was {split} but only valid "
                f"splits are: {VALID_SPLITS}"
            )
        self.split = split
        self.transforms = transforms
        self.catalog = _get_split(
            split=split,
            catalog=self.catalog,
            train_percentage=train_split_ratio,
            random_seed=random_seed,
        )

    def __getitem__(self, index):
        """
        Get the image and corresponding bounding boxes for that index
        Args:
            index:

        Returns (Tuple(Image,List(BBox2D))): Tuple comprising the image and
        bounding boxes found in that image with transforms applied.

        """
        cap = self.catalog.iloc[index]
        capture_file = cap.filename
        ann = cap["annotation.values"]

        capture = Image.open(os.path.join(self._data_path, capture_file))
        capture = capture.convert("RGB")  # Remove alpha channel
        annotation = read_bounding_box_2d(ann, self.label_mappings)

        if self.transforms:
            capture, annotation = self.transforms(capture, annotation)

        return capture, annotation

    def __len__(self):
        return len(self.catalog)

    def _cleanup(self, catalog):
        """
        remove rows with captures that having missing files and remove examples
        which have no annotations i.e. an image without any objects
        Args:
            catalog (pandas dataframe):

        Returns: dataframe without rows corresponding to captures that have
        missing files and removes examples which have no annotations i.e. an
        image without any objects.

        """
        catalog = self._remove_captures_with_missing_files(
            self._data_path, catalog
        )
        catalog = self._remove_captures_without_bboxes(catalog)

        return catalog

    @staticmethod
    def _remove_captures_without_bboxes(catalog):
        """Remove captures without bounding boxes from catalog

        Args:
            catalog (pd.Dataframe): The loaded catalog of the dataset

        Returns:
            A pandas dataframe with empty bounding boxes removed
        """
        keep_mask = catalog["annotation.values"].apply(len) > 0

        return catalog[keep_mask]

    @staticmethod
    def _remove_captures_with_missing_files(root, catalog):
        """Remove captures where image files are missing

        During the synthetic dataset download process, some of the files might
        be missing due to temporary http request issues or url corruption.
        We should remove these captures from catalog so that it does not
        stop the training pipeline.

        Args:
            catalog (pd.Dataframe): The loaded catalog of the dataset

        Returns:
            A pandas dataframe of the catalog with missing files removed
        """

        def exists(capture_file):
            path = Path(root) / capture_file

            return path.exists()

        keep_mask = catalog.filename.apply(exists)

        return catalog[keep_mask]

    @staticmethod
    def _preprocess_dataset(data_path):
        """ Preprocess dataset inside data_path and un-archive if necessary.

        Args:
            data_path (str): Path where dataset is stored.

        Return:
            Path of the dataset files.
        """

        # This is for dataset generated by unity simulation.
        # In this case, all data are downloaded directly in the data_path
        if SynDetection2D.is_dataset_files_present(data_path):
            return data_path

        # This is for dataset stored in a single zip file.
        # In this case, all data are downloaded to a single zip file under
        # data_path. We will un-archive this path and change the data_path
        # to the un-archived folder
        archive_file = Path(data_path) / SynDetection2D.ARCHIVE_FILE
        if not archive_file.exists():
            raise DatasetNotFoundError(
                f"Expecting a file {archive_file} under {data_path} or files "
                f"directly exist under {data_path}"
            )

        unarchived_path = Path(data_path) / SynDetection2D.SUBFOLDER
        shutil.unpack_archive(
            filename=archive_file, extract_dir=unarchived_path
        )

        return unarchived_path

    @staticmethod
    def is_dataset_files_present(data_path):
        return os.path.isdir(data_path) and any(glob.glob(f"{data_path}/**/*"))



import copy
import logging

import numpy as np
import torch
import torchvision
from ignite.metrics import Loss
from torchvision import transforms as T
from torchvision.transforms import functional as F

from datasetinsights.datasets import Dataset
from datasetinsights.evaluation_metrics import EvaluationMetric
from datasetinsights.io.loader import create_loader
from datasetinsights.io.transforms import Compose, RandomHorizontalFlip
from datasetinsights.stats.visualization.plots import decode_segmap, grid_plot

from .base import Estimator

logger = logging.getLogger(__name__)

# Normalization constants (heuristics) from ImageNet dataset
_IMGNET_MEAN = (0.485, 0.456, 0.406)
_IMGNET_STD = (0.229, 0.224, 0.225)

# Inverse Normalization constants
_INV_IMGNET_MEAN = (-0.485 / 0.229, -0.456 / 0.224, -0.406 / 0.225)
_INV_IMGNET_STD = (1.0 / 0.229, 1.0 / 0.224, 1.0 / 0.5)


def pad_if_smaller(img, size, fill=0):
    min_size = min(img.size)
    if min_size < size:
        ow, oh = img.size
        padh = size - oh if oh < size else 0
        padw = size - ow if ow < size else 0
        img = F.pad(img, (0, 0, padw, padh), fill=fill)

    return img


class RandomCrop:
    def __init__(self, size):
        self.size = size

    def __call__(self, image, target):
        image = pad_if_smaller(image, self.size)
        target = pad_if_smaller(target, self.size, fill=255)
        crop_params = T.RandomCrop.get_params(image, (self.size, self.size))
        image = F.crop(image, *crop_params)
        target = F.crop(target, *crop_params)

        return image, target


class ToTensor:
    """Convert a pair of (image, target) to tensor
    """

    def __call__(self, image, target):
        image = F.to_tensor(image)
        target = torch.as_tensor(np.asarray(target), dtype=torch.int64)

        return image, target


class Normalize:
    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, image, target):
        image = F.normalize(image, mean=self.mean, std=self.std)

        return image, target


class DeeplabV3(Estimator):
    """ DeeplabV3 Model https://arxiv.org/abs/1706.05587

    Args:
        config (CfgNode): estimator config
        writer: Tensorboard writer object
        checkpointer: Model checkpointer callback to save models
        device: model training on device (cpu|cuda)
    Attributes:
        backbone: model backbone (resnet50|resnet101)
        num_classes: number of classes for semantic segmentation
        model: tensorflow or pytorch graph
        writer: Tensorboard writer object
        checkpointer: Model checkpointer callback to save models
        device: model training on device (cpu|cuda)
        optimizer: pytorch optimizer
        lr_scheduler: pytorch learning rate scheduler
    """

    def __init__(
        self,
        *,
        config,
        writer,
        checkpointer,
        device,
        checkpoint_file=None,
        **kwargs,
    ):
        self.config = config
        self.writer = writer
        self.checkpointer = checkpointer
        self.device = device

        self.backbone = config.backbone
        self.num_classes = config.num_classes

        model_name = "deeplabv3_" + config.backbone
        self.model = torchvision.models.segmentation.__dict__[model_name](
            num_classes=self.num_classes
        )

        opname = self.config.optimizer.name
        if opname == "Adam":
            optimizer = torch.optim.Adam(
                self.model.parameters(), **self.config.optimizer.args
            )

            # use fixed learning rate when using Adam
            lr_scheduler = torch.optim.lr_scheduler.LambdaLR(
                optimizer, lambda x: 1.0
            )
        else:
            raise ValueError(f"Unsupported optimizer type {opname}")

        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler

        # load estimators from file if checkpoint_file exists
        if checkpoint_file:
            self.checkpointer.load(self, checkpoint_file)

    @staticmethod
    def _transforms(is_train=True, crop_size=769):
        """Transformations for a pair of input and target image

        Args:
            is_train (bool): indicator whether this is a transformation
                during training (default: True)
            crop_size (int): crop size. Images will be cropped to
                (crop_size, crop_size)
        """
        transforms = []
        if is_train:
            transforms.append(RandomHorizontalFlip(0.5))
            transforms.append(RandomCrop(crop_size))
        transforms.append(ToTensor())
        transforms.append(Normalize(mean=_IMGNET_MEAN, std=_IMGNET_STD))

        return Compose(transforms)

    @staticmethod
    def _loss_fn(outputs, target):
        """ Compute loss

        Args:
            outputs (dict): named output of deeplabv3 model. Since this
            implementation outpus two semantic segmentation images from two
            heads of the model, we are expecting dict of tow keys
            "out" and "aux" that corresponds to two pytorch tenor of images.
            target (torch.Tensor): ground truth 2D image tensor

        Returns:
            numerical value of loss
        """
        losses = {}
        for name, x in outputs.items():
            losses[name] = torch.nn.functional.cross_entropy(
                x, target, ignore_index=255
            )

        if len(losses) == 1:
            return losses["out"]

        return losses["out"] + 0.5 * losses["aux"]

    def _train_one_epoch(self, loader, epoch):
        """ Train one epoch

        Args:
            loader (DataLoader): pytorch dataloader
            epoch (int): the current epoch number
        """
        logger.info(f"Epoch[{epoch}] training started.")
        self.model.train()
        n_batch = len(loader)
        accumulation_steps = self.config.train.accumulation_steps
        loss_metric = Loss(self._loss_fn)

        self.optimizer.zero_grad()
        for i, (image, target) in enumerate(loader):
            image, target = image.to(self.device), target.to(self.device)
            output = self.model(image)
            loss = self._loss_fn(output, target)
            loss.backward()

            # Accumulated Gradients are only updated after X steps.
            # This creates an effective batch size of
            # batch_size * accumulation_steps
            if (i + 1) % accumulation_steps == 0:
                self.optimizer.step()
                self.lr_scheduler.step()
                self.optimizer.zero_grad()

            loss_metric.update((output, target))

            iter_num = (i + 1) % n_batch
            logger.debug(
                f"Epoch[{epoch}] Iteration[{iter_num}/{n_batch}] "
                f"Loss: {loss:.3f}"
            )

        epoch_loss = loss_metric.compute()
        logger.info(
            f"Epoch[{epoch}] training completed. Loss: {epoch_loss:.3f}"
        )
        self.writer.add_scalar("training/loss", epoch_loss, epoch)

        loss_metric.reset()

    def _evaluate_one_epoch(self, loader, epoch):
        """ Evaluate one epoch

        Args:
            loader (DataLoader): pytorch dataloader
            epoch (int): the current epoch number
        """
        logger.info(f"Epoch[{epoch}] evaluation started")
        self.model.eval()
        loss_metric = Loss(self._loss_fn)

        # TODO: Support other metrics other than IoU and support multiple
        # mettics
        iou_metric = EvaluationMetric.create(
            self.config.metric, num_classes=self.num_classes
        )
        with torch.no_grad():
            for image, target in loader:
                image, target = image.to(self.device), target.to(self.device)
                output = self.model(image)

                loss_metric.update((output, target))
                iou_metric.update((output["out"], target))

        loss = loss_metric.compute()
        iou = iou_metric.compute()

        # some classes are not used in cityscapes evaluation.
        # TODO: Move class masking logic to IoU metric class.
        keep_mask = [
            not c.ignore_in_eval
            for c in torchvision.datasets.Cityscapes.classes
        ]
        class_names = [c.name for c in torchvision.datasets.Cityscapes.classes]
        iou_info = {
            name: f"{iou[i].item():.3f}"
            for i, name in enumerate(class_names)
            if keep_mask[i]
        }
        miou = iou[keep_mask].mean()

        logger.info(
            f"Epoch[{epoch}] evaluation completed. "
            f"Loss: {loss:.3f}, mIoU: {miou:.3f}\n"
            f"IoU per class: {iou_info}"
        )
        self.writer.add_scalar("validation/loss", loss, epoch)
        self.writer.add_scalar("validation/miou", miou, epoch)

        inv_normalize = T.Normalize(mean=_INV_IMGNET_MEAN, std=_INV_IMGNET_STD)
        # Visualize segmentation images from last mini-batch
        n_images = list(image.shape)[0]
        image_grid = []
        for i in range(n_images):
            img = inv_normalize(image[i, :]).permute(1, 2, 0).cpu().numpy()
            out = decode_segmap(output["out"][i, :].max(0)[1].cpu().numpy())
            tgt = decode_segmap(target[i, :].cpu().numpy())
            image_grid.append([img, out, tgt])

        fig = grid_plot(image_grid)
        self.writer.add_figure("validation/visualize", fig, epoch)

        loss_metric.reset()
        iou_metric.reset()

    def train(self, **kwargs):
        config = self.config
        train_dataset = Dataset.create(
            config.train.dataset,
            split="train",
            data_root=config.system.data_root,
            transforms=self._transforms(
                is_train=True, crop_size=config.train.crop_size
            ),
        )
        train_loader = create_loader(
            train_dataset,
            batch_size=config.train.batch_size,
            num_workers=config.system.workers,
            dryrun=config.system.dryrun,
        )

        val_dataset = Dataset.create(
            config.val.dataset,
            split="val",
            data_root=config.system.data_root,
            transforms=self._transforms(is_train=False),
        )
        val_loader = create_loader(
            val_dataset,
            batch_size=config.val.batch_size,
            num_workers=config.system.workers,
            dryrun=config.system.dryrun,
        )

        logger.info("Start training estimator: %s", type(self).__name__)
        self.model.to(self.device)
        n_epochs = config.train.epochs
        val_interval = config.system.val_interval
        for epoch in range(1, n_epochs + 1):
            logger.info(f"Training Epoch[{epoch}/{n_epochs}]")
            self._train_one_epoch(train_loader, epoch)

            if epoch % val_interval == 0:
                self._evaluate_one_epoch(val_loader, epoch)

            self.checkpointer.save(self, epoch=epoch)

        self.writer.close()

    def evaluate(self, **kwargs):
        config = self.config
        test_dataset = Dataset.create(
            config.test.dataset,
            split="test",
            data_root=config.system.data_root,
            transforms=self._transforms(is_train=False),
        )
        test_loader = create_loader(
            test_dataset,
            batch_size=config.test.batch_size,
            num_workers=config.system.workers,
            dryrun=config.system.dryrun,
        )

        logger.info("Start evaluating estimator: %s", type(self).__name__)
        self.model.to(self.device)
        self._evaluate_one_epoch(test_loader, epoch=1)
        self.writer.close()

    def save(self, path):
        """ Serialize Estimator to path

        Args:
            path (str): full path to save serialized estimator

        Returns:
            saved full path of the serialized estimator
        """
        save_dict = {"model": self.model.state_dict(), "config": self.config}
        torch.save(save_dict, path)

        return path

    def load(self, path):
        """ Load Estimator from path

        Args:
            path (str): full path to the serialized estimator
        """
        checkpoint = torch.load(path)
        self.model.load_state_dict(checkpoint["model"])

        loaded_config = copy.deepcopy(checkpoint["config"])
        stored_config = copy.deepcopy(self.config)
        del stored_config["checkpoint_file"]
        del loaded_config["checkpoint_file"]
        if stored_config != loaded_config:
            logger.warning(
                f"Found difference in estimator config."
                f"Estimator loaded from {path} was trained using "
                f"config: "
                f"{loaded_config}. However, the current config is: "
                f"{self.config}."
            )



import json
import logging
import os

logger = logging.getLogger(__name__)


class KubeflowPipelineWriter(object):
    """
    Serializes metrics dictionary genereated during model training/evaluation to
    JSON and store in a file.

    Args:
        filename (str): Name of the file to which the writer will save metrics
        filepath (str): Path where the file will be stored

    Attributes:
        filename (str): Name of the file to which the writer will save metrics
        filepath (str): Path where the file will be stored
        data_dict (dict): A dictionary to save metrics name and value pairs
        data: Dictionary to be JSON serialized
    """

    def __init__(self, filename="mlpipeline-metrics.json", filepath="/"):
        """
        Creates KubeflowPipelineWriter that will write out metrics to the output
        file
        """

        self.filename = filename
        self.filepath = filepath
        self.data_dict = {}
        self.data = {"metrics": []}

    def add_metric(self, name, val):
        """
        Adds metric to the data dictionary of the writer

        Note: Using same name key will overwrite the previous value as the
        current strategy is to save only the metrics generated in last epoch

        Args:
            name (str): Name of the metric
            val (float): Value of the metric
        """

        logger.debug("Metric {0} with value: {1} added".format(name, val))
        self.data_dict[name] = val

    def write_metric(self):
        """
        Saves all the metrics added previously to a file in the format required
        by kubeflow
        """
        if not self.data_dict:
            logger.warning("No metrics generated to be saved.")
            return

        for key, val in self.data_dict.items():
            self.data["metrics"].append(
                {"name": key, "numberValue": val, "format": "RAW"}
            )
        if not os.path.exists(self.filepath):
            os.makedirs(self.filepath)
        with open(os.path.join(self.filepath, self.filename), "w") as f:
            json.dump(self.data, f)

        logger.debug(
            f"Metrics file {self.filename} saved at path:" f" {self.filepath}"
        )



import copy
import logging
import random
from itertools import permutations

import kornia
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.models as models
from ignite.metrics import Loss
from PIL import Image
from torchvision.transforms import Compose
from torchvision.transforms.functional import to_tensor

from datasetinsights.datasets import Dataset
from datasetinsights.evaluation_metrics import EvaluationMetric
from datasetinsights.io.loader import create_loader
from datasetinsights.io.transforms import RandomHorizontalFlip, Resize
from datasetinsights.stats.visualization.plots import grid_plot

from .base import Estimator

logger = logging.getLogger(__name__)
logging.getLogger("matplotlib.font_manager").setLevel(logging.WARNING)


class RandomChannelSwap:
    """Swap color channel of the image

    Args:
        probability: the probability to swap color channel of the image
    """

    def __init__(self, probability):
        self.probability = probability

    def __call__(self, sample):
        image, depth = sample

        if random.random() < self.probability:
            image = np.asarray(image)
            perm = random.choice(list(permutations(range(3))))
            image = Image.fromarray(image[..., perm])

        return (image, depth)


class ToTensor(object):
    """Convert the image and depth to tensor."""

    def __call__(self, sample):
        image, depth = sample
        image = to_tensor(image)
        depth = to_tensor(depth)

        return (image, depth)


class UpSample(nn.Sequential):
    def __init__(self, skip_input, output_features):
        super().__init__()
        self.convA = nn.Conv2d(
            skip_input, output_features, kernel_size=3, stride=1, padding=1
        )
        self.leakyreluA = nn.LeakyReLU(0.2)
        self.convB = nn.Conv2d(
            output_features, output_features, kernel_size=3, stride=1, padding=1
        )
        self.leakyreluB = nn.LeakyReLU(0.2)

    def forward(self, x, concat_with):
        up_x = F.interpolate(
            x,
            size=[concat_with.size(2), concat_with.size(3)],
            mode="bilinear",
            align_corners=True,
        )
        return self.leakyreluB(
            self.convB(
                self.leakyreluA(
                    self.convA(torch.cat([up_x, concat_with], dim=1))
                )
            )
        )


class Decoder(nn.Module):
    def __init__(self, num_features=2208, decoder_width=0.5):
        super().__init__()
        features = int(num_features * decoder_width)

        self.conv2 = nn.Conv2d(
            num_features, features, kernel_size=1, stride=1, padding=1
        )

        self.up1 = UpSample(
            skip_input=features // 1 + 384, output_features=features // 2
        )
        self.up2 = UpSample(
            skip_input=features // 2 + 192, output_features=features // 4
        )
        self.up3 = UpSample(
            skip_input=features // 4 + 96, output_features=features // 8
        )
        self.up4 = UpSample(
            skip_input=features // 8 + 96, output_features=features // 16
        )

        self.conv3 = nn.Conv2d(
            features // 16, 1, kernel_size=3, stride=1, padding=1
        )

    def forward(self, features):
        x_block0 = features[3]
        x_block1 = features[4]
        x_block2 = features[6]
        x_block3 = features[8]
        x_block4 = features[11]
        x_d0 = self.conv2(x_block4)
        x_d1 = self.up1(x_d0, x_block3)
        x_d2 = self.up2(x_d1, x_block2)
        x_d3 = self.up3(x_d2, x_block1)
        x_d4 = self.up4(x_d3, x_block0)
        return self.conv3(x_d4)


class Encoder(nn.Module):
    def __init__(self):
        super().__init__()
        self.original_model = models.densenet161(pretrained=True)

    def forward(self, x):
        features = [x]
        for k, v in self.original_model.features._modules.items():
            features.append(v(features[-1]))
        return features


class DenseDepthModel(nn.Module):
    def __init__(self):
        super().__init__()
        self.encoder = Encoder()
        self.decoder = Decoder()

    def forward(self, x):
        return self.decoder(self.encoder(x))


class DenseDepth(Estimator):
    """
    Attributes:
        config: estimator config
        writer: Tensorboard writer object
        model: tensorflow or pytorch graph
        checkpointer: Model checkpointer callback to save models
        device: model training on device (cpu|cuda)
        optimizer: pytorch optimizer
    """

    def __init__(
        self, config, writer, checkpointer, device, checkpoint_file, **kwargs
    ):
        """
        Args:
        config: estimator config
        writer: Tensorboard writer object
        checkpointer: Model checkpointer callback to save models
        device: model training on device (cpu|cuda)
        """
        # initialize new model or load pre-trained model
        self.config = config
        self.checkpointer = checkpointer
        self.device = device
        self.writer = writer

        self.model = DenseDepthModel()
        logger.info("DenseDepth model is created.")

        opname = self.config.optimizer.name

        if opname == "Adam":
            optimizer = torch.optim.Adam(
                self.model.parameters(), self.config.optimizer.lr
            )
        else:
            raise ValueError(f"Unsupported optimizer type {opname}")

        self.optimizer = optimizer

        # load estimators from file if checkpoint_file exists
        if checkpoint_file:
            self.checkpointer.load(self, checkpoint_file)

    @staticmethod
    def _NYU_transforms(is_train=True):
        transforms = []
        if is_train:
            transforms.append(RandomHorizontalFlip(0.5))
            transforms.append(RandomChannelSwap(0.25))
        transforms.append(Resize(target_size=(240)))
        transforms.append(ToTensor())
        return Compose(transforms)

    @staticmethod
    def _loss_fn(output, depth_n):
        """ Compute loss

        Args:
            output (torch.Tensor): predicted 2D tensor
            depth_n (torch.Tensor): ground truth 2D depth tensor

        Returns:
            numerical value of loss
        """
        _, _, width, height = output.size()
        l1_criterion = nn.L1Loss()
        l_depth = l1_criterion(output, depth_n)
        #  window size = 11 from original repo
        ssim_criterion = kornia.losses.SSIM(min(11, width, height))
        l_ssim = ssim_criterion(output, depth_n).mean().item()
        loss = (1.0 * l_ssim) + (0.1 * l_depth)

        return loss

    def _evaluate_one_epoch(self, loader, epoch, n_epochs):
        """ Evaluate one epoch

        Args:
            loader (DataLoader): pytorch dataloader
            epoch (int): the current epoch number
            n_epochs (int): total epoch number
        """
        logger.info(f"Epoch[{epoch}/{n_epochs}] evaluation started")
        self.model.eval()
        loss_metric = Loss(self._loss_fn)

        # metrics
        # The RMSE is smaller than the normal scale. Because we didn't do the:
        # depth = self.to_tensor(depth).float() * 10
        are_metric = EvaluationMetric.create("AverageRelativeError")
        log10_metric = EvaluationMetric.create("AverageLog10Error")
        rmse_metric = EvaluationMetric.create("RootMeanSquareError")
        a1_metric = EvaluationMetric.create("ThresholdAccuracy", threshold=1.25)
        a2_metric = EvaluationMetric.create(
            "ThresholdAccuracy", threshold=1.25 ** 2
        )
        a3_metric = EvaluationMetric.create(
            "ThresholdAccuracy", threshold=1.25 ** 3
        )

        with torch.no_grad():
            for image, depth in loader:

                image = image.to(self.device)
                depth_n = depth.to(self.device)

                output = self.model(image)

                loss_metric.update((output, depth_n))
                depth_image_pair = (output.cpu().numpy(), depth_n.cpu().numpy())
                are_metric.update(depth_image_pair)
                log10_metric.update(depth_image_pair)
                rmse_metric.update(depth_image_pair)
                a1_metric.update(depth_image_pair)
                a2_metric.update(depth_image_pair)
                a3_metric.update(depth_image_pair)

        # Compute the loss
        loss = loss_metric.compute()
        are_metric_val = are_metric.compute()
        log10_metric_val = log10_metric.compute()
        rmse_metric_val = rmse_metric.compute()
        a1_metric_val = a1_metric.compute()
        a2_metric_val = a2_metric.compute()
        a3_metric_val = a3_metric.compute()

        logger.info(
            f"Epoch[{epoch}/{n_epochs}] evaluation completed.\n"
            f"Validation Loss: {loss:.3f}\n"
            f"Average Relative Error: {are_metric_val:.3f}\n"
            f"Average Log10 Error: {log10_metric_val:.3f}\n"
            f"Root Mean Square Error: {rmse_metric_val:.3f}\n"
            f"Threshold Accuracy (delta1): {a1_metric_val:.3f}\n"
            f"Threshold Accuracy (delta2): {a2_metric_val:.3f}\n"
            f"Threshold Accuracy (delta3): {a3_metric_val:.3f}\n"
        )

        self.writer.add_scalar("Validation/loss", loss, epoch)
        self.writer.add_scalar(
            "Validation/Average_Relative_Error", are_metric_val, epoch
        )
        self.writer.add_scalar(
            "Validation/Average_Log10_Error", log10_metric_val, epoch
        )
        self.writer.add_scalar(
            "Validation/Root_Mean_Square_Error", rmse_metric_val, epoch
        )
        self.writer.add_scalar(
            "Validation/Threshold_Accuracy__delta1_", a1_metric_val, epoch
        )
        self.writer.add_scalar(
            "Validation/Threshold_Accuracy__delta2_", a2_metric_val, epoch
        )
        self.writer.add_scalar(
            "Validation/Threshold_Accuracy__delta3_", a3_metric_val, epoch
        )

        # Visualize depth images from last mini-batch
        n_images = image.shape[0]
        image_grid = []
        gray_depth_grid = []
        for i in range(n_images):
            rgb_image = image[i].permute(1, 2, 0)
            image_grid.append([rgb_image.cpu().numpy()])
            gray_depth_grid.append(
                [
                    output[i].permute(1, 2, 0)[:, :, -1].cpu().numpy(),
                    depth_n[i].permute(1, 2, 0)[:, :, -1].cpu().numpy(),
                ]
            )
        # Add figures
        fig = grid_plot(image_grid)
        gray_figs = grid_plot(gray_depth_grid, img_type="gray")
        self.writer.add_figure("Validation/visualize_depths", gray_figs, epoch)
        self.writer.add_figure("Validation/visualize_image", fig, epoch)

        loss_metric.reset()
        are_metric.reset()
        log10_metric.reset()
        rmse_metric.reset()
        a1_metric.reset()
        a2_metric.reset()
        a3_metric.reset()

    def train(self, **kwargs):
        # Training parameters
        config = self.config
        optimizer = self.optimizer
        val_interval = config.system.val_interval
        writer = self.writer

        # Load data
        train_dataset = Dataset.create(
            config.train.dataset,
            split="train",
            data_root=config.system.data_root,
            transforms=self._NYU_transforms(is_train=True),
        )

        train_loader = create_loader(
            train_dataset,
            batch_size=config.train.batch_size,
            num_workers=config.system.workers,
            dryrun=config.system.dryrun,
        )

        val_dataset = Dataset.create(
            config.val.dataset,
            split="test",
            data_root=config.system.data_root,
            transforms=self._NYU_transforms(is_train=False),
        )

        val_loader = create_loader(
            val_dataset,
            batch_size=config.val.batch_size,
            num_workers=config.system.workers,
            dryrun=config.system.dryrun,
        )

        # Logging
        logger.info("Start training estimator: %s", type(self).__name__)

        self.model.to(self.device)
        n_epochs = config.train.epochs

        # Start training
        for epoch in range(1, n_epochs + 1):
            logger.info(f"Epoch[{epoch}/{n_epochs}] training started.")
            loss_metric = Loss(self._loss_fn)
            self.model.train()
            N = len(train_loader)
            accumulation_steps = self.config.train.accumulation_steps
            optimizer.zero_grad()
            for i, (image, depth) in enumerate(train_loader):
                # Prepare sample and depth
                image = image.to(self.device)
                depth_n = depth.to(self.device)

                # Predict
                output = self.model(image)

                # Compute loss
                loss = self._loss_fn(output, depth_n)

                # Backward
                loss.backward()

                if (i + 1) % accumulation_steps == 0:
                    self.optimizer.step()
                    self.optimizer.zero_grad()

                loss_metric.update((output, depth_n))

                # Log progress
                logger.debug(f"[{i}/{N}] Loss: {loss:.4f}")

            epoch_loss = loss_metric.compute()
            if epoch % val_interval == 0:
                self._evaluate_one_epoch(val_loader, epoch, n_epochs)

            # Record epoch's intermediate results
            writer.add_scalar("Training/Loss", epoch_loss, epoch)
            self.checkpointer.save(self, epoch=epoch)

        self.writer.close()

    def evaluate(self, **kwargs):
        config = self.config
        test_dataset = Dataset.create(
            config.test.dataset,
            split="test",
            data_root=config.system.data_root,
            transforms=self._NYU_transforms(is_train=False),
        )
        test_loader = create_loader(
            test_dataset,
            batch_size=config.test.batch_size,
            num_workers=config.system.workers,
            dryrun=config.system.dryrun,
        )

        logger.info("Start evaluating estimator: %s", type(self).__name__)
        self.model.to(self.device)
        self._evaluate_one_epoch(test_loader, 1, 1)
        self.writer.close()

    def save(self, path):
        """ Serialize Estimator to path

        Args:
            path (str): full path to save serialized estimator

        Returns:
            saved full path of the serialized estimator
        """
        save_dict = {"model": self.model.state_dict(), "config": self.config}
        torch.save(save_dict, path)

        return path

    def load(self, path):
        """ Load Estimator from path

        Args:
            path (str): full path to the serialized estimator
        """
        checkpoint = torch.load(path)
        self.model.load_state_dict(checkpoint["model"])

        loaded_config = copy.deepcopy(checkpoint["config"])
        stored_config = copy.deepcopy(self.config)
        del stored_config["checkpoint_file"]
        del loaded_config["checkpoint_file"]
        if stored_config != loaded_config:
            logger.warning(
                f"Found difference in estimator config."
                f"Estimator loaded from {path} was trained using config: "
                f"{loaded_config}. However, the current config is: "
                f"{self.config}."
            )



from .base import Dataset
from .cityscapes import Cityscapes
from .coco import CocoDetection
from .groceries_real import GoogleGroceriesReal, GroceriesReal
from .nyudepth import NyuDepth
from .synthetic import SynDetection2D

__all__ = [
    "Cityscapes",
    "CocoDetection",
    "Dataset",
    "GroceriesReal",
    "GoogleGroceriesReal",
    "NyuDepth",
    "SynDetection2D",
]



r"""Reference.

http://cocodataset.org/#detection-eval
https://arxiv.org/pdf/1502.05082.pdf
https://github.com/rafaelpadilla/Object-Detection-Metrics/issues/22
"""
import collections

import numpy as np

from datasetinsights.io.bbox import group_bbox2d_per_label

from .base import EvaluationMetric
from .records import Records


class AverageRecall(EvaluationMetric):
    """2D Bounding Box Average Recall metrics.

    This metric would calculate average recall (AR) for each label under
    an iou threshold (default: 0.5). The maximum number of detections
    per image is limited (default: 100).

    Args:
        iou_threshold (float): iou threshold (default: 0.5)
        max_detections (int): max detections per image (default: 100)
    """

    TYPE = "metric_per_label"

    def __init__(self, iou_threshold=0.5, max_detections=100):
        self._label_records = collections.defaultdict(
            lambda: Records(iou_threshold=self._iou_threshold)
        )
        self._gt_bboxes_count = collections.defaultdict(int)
        self._iou_threshold = iou_threshold
        self._max_detections = max_detections

    def reset(self):
        """Reset AR metrics."""
        self._label_records = collections.defaultdict(
            lambda: Records(iou_threshold=self._iou_threshold)
        )
        self._gt_bboxes_count = collections.defaultdict(int)

    def update(self, mini_batch):
        """Update records per mini batch.

        Args:
            mini_batch (list(list)): a list which contains batch_size of
            gt bboxes and pred bboxes pair in each image.
            For example, if batch size = 2, mini_batch looks like:
            [[gt_bboxes1, pred_bboxes1], [gt_bboxes2, pred_bboxes2]]
            where gt_bboxes1, pred_bboxes1 contain gt bboxes and pred bboxes
            in one image
        """
        for bboxes in mini_batch:
            gt_bboxes, pred_bboxes = bboxes
            for gt_bbox in gt_bboxes:
                self._gt_bboxes_count[gt_bbox.label] += 1

            pred_bboxes = sorted(
                pred_bboxes, key=lambda bbox: bbox.score, reverse=True
            )
            if len(pred_bboxes) > self._max_detections:
                pred_bboxes = pred_bboxes[: self._max_detections]
            bboxes_per_label = group_bbox2d_per_label(pred_bboxes)
            for label in bboxes_per_label:
                self._label_records[label].add_records(
                    gt_bboxes, bboxes_per_label[label]
                )

    def compute(self):
        """Compute AR for each label.

        Returns:
            dict: a dictionary of AR scores per label.
        """
        average_recall = {}
        _label_records = self._label_records
        for label in self._gt_bboxes_count:
            # if there are no predicted boxes with this label
            if label not in _label_records:
                average_recall[label] = 0
                continue

            match_results = _label_records[label].match_results
            _gt_bboxes_count = self._gt_bboxes_count[label]

            # The number of TP
            sum_tp = sum(list(zip(*match_results))[1])

            max_recall = sum_tp / _gt_bboxes_count

            average_recall[label] = max_recall

        return average_recall


class MeanAverageRecallAverageOverIOU(EvaluationMetric):
    """2D Bounding Box Mean Average Recall metrics.

    This implementation computes Mean Average Recall (mAR) metric,
    which is implemented as the Average Recall average over all
    labels and IOU = 0.5:0.95:0.05. The max detections
    per image is limited to 100.

    .. math:: mAR^{IoU=0.5:0.95:0.05} = mean_{label,IoU}
    AR^{label, IoU=0.5:0.95:0.05}
    """

    TYPE = "scalar"

    IOU_THRESHOULDS = np.linspace(
        0.5, 0.95, np.round((0.95 - 0.5) / 0.05) + 1, endpoint=True
    )

    def __init__(self):
        self._mar_per_iou = [
            AverageRecall(iou)
            for iou in MeanAverageRecallAverageOverIOU.IOU_THRESHOULDS
        ]

    def reset(self):
        [mean_ar.reset() for mean_ar in self._mar_per_iou]

    def update(self, mini_batch):
        for mean_ar in self._mar_per_iou:
            mean_ar.update(mini_batch)

    def compute(self):
        """Compute mAR over IOU.
        """
        result = np.mean(
            [
                value
                for ar in self._mar_per_iou
                for value in ar.compute().values()
            ]
        )
        return result



"""faster rcnn pytorch train and evaluate."""

import copy
import logging
import math
import os
from typing import Dict, List, Tuple

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.distributed as dist
import torchvision
from codetiming import Timer
from tensorboardX import SummaryWriter
from torchvision import transforms
from tqdm import tqdm

import datasetinsights.constants as const
from datasetinsights.datasets import Dataset
from datasetinsights.evaluation_metrics.base import EvaluationMetric
from datasetinsights.io.bbox import BBox2D
from datasetinsights.io.transforms import Compose
from datasetinsights.torch_distributed import get_world_size, is_master

from .base import Estimator

MAX_BOXES_PER_IMAGE = 100
logger = logging.getLogger(__name__)
DEFAULT_ACCUMULATION_STEPS = 1
TRAIN = "train"
VAL = "val"
TEST = "test"


class FasterRCNN(Estimator):
    """
    Faster-RCNN train/evaluate implementation for object detection.

    https://github.com/pytorch/vision/tree/master/references/detection
    https://arxiv.org/abs/1506.01497
    Args:
        config (CfgNode): estimator config
        box_score_thresh: (optional) default threshold is 0.05
        distributed: whether or not the estimator is distributed
        kfp_metrics_filename: Kubeflow Metrics filename
        kfp_metrics_dir: Path to the directory where Kubeflow
         metrics files are stored


    https://github.com/pytorch/vision/tree/master/references/detection
    https://arxiv.org/abs/1506.01497

    Attributes:
        model: pytorch model
        writer: Tensorboard writer object
        kfp_writer: KubeflowPipelineWriter object
        checkpointer: Model checkpointer callback to save models
        device: model training on device (cpu|cuda)
    """

    def __init__(
        self,
        *,
        config,
        writer,
        kfp_writer,
        checkpointer,
        box_score_thresh=0.05,
        no_cuda=None,
        checkpoint_file=None,
        **kwargs,
    ):
        """initiate estimator."""

        logger.info(f"initializing faster rcnn")
        self.config = config

        self._init_distributed_mode()
        self.no_cuda = no_cuda
        self._init_device()
        self.writer = SummaryWriter(writer.logdir, write_to_disk=is_master())

        self.kfp_writer = kfp_writer
        checkpointer.distributed = self.distributed
        self.checkpointer = checkpointer

        model_name = f"fasterrcnn_{self.config.backbone}_fpn"
        self.model = torchvision.models.detection.__dict__[model_name](
            num_classes=config.num_classes,
            pretrained_backbone=config.pretrained_backbone,
            pretrained=config.pretrained,
            box_detections_per_img=MAX_BOXES_PER_IMAGE,
            box_score_thresh=box_score_thresh,
        )
        self.model_without_ddp = self.model
        self.sync_metrics = config.get("synchronize_metrics", True)
        self.metrics = {}
        for metric_key, metric in config.metrics.items():
            self.metrics[metric_key] = EvaluationMetric.create(metric.name)
        self.model.to(self.device)

        if self.distributed:
            self.model = torch.nn.parallel.DistributedDataParallel(
                self.model, device_ids=[self.gpu]
            )
            self.model_without_ddp = self.model.module

        if checkpoint_file:
            self.checkpointer.load(self, checkpoint_file)

    def _init_distributed_mode(self):
        if "RANK" in os.environ and "WORLD_SIZE" in os.environ:
            logger.info(f"found RANK and WORLD_SIZE in environment")
            self.rank = int(os.environ["RANK"])
            self.world_size = int(os.environ["WORLD_SIZE"])
            self.gpu = int(os.environ["LOCAL_RANK"])
        elif "SLURM_PROCID" in os.environ:
            logger.info(f"found 'SLURM_PROCID' in environment")
            self.rank = int(os.environ["SLURM_PROCID"])
            self.gpu = self.rank % torch.cuda.device_count()
        else:
            self.gpu = 0
            self.rank = 0
            logger.info("Not using distributed mode")
            self.distributed = False

        if self.gpu != 0 and self.rank != 0:
            device_count = torch.cuda.device_count()
            logger.info(f"device count: {torch.cuda.device_count()}")
            logger.info(f"world size: {self.world_size}")
            logger.info(f"gpu: {self.gpu}")
            logger.info(f"local rank {self.rank}")
            if device_count == 0:
                logger.info("No cuda devices found, will not parallelize")
                self.distributed = False
            else:
                if not is_master():
                    logging.disable(logging.ERROR)
                self.distributed = True
                torch.cuda.set_device(self.gpu)

                torch.distributed.init_process_group(
                    backend="nccl",
                    init_method="env://",
                    world_size=self.world_size,
                    rank=self.rank,
                )
                torch.distributed.barrier()

    def _init_device(self):
        if torch.cuda.is_available() and not self.no_cuda:
            self.device = torch.device("cuda")
        else:
            self.device = torch.device("cpu")

    def train(self, train_data, val_data=None, **kwargs):
        """start training, save trained model per epoch.

        Args:
            train_data: Directory on localhost where train dataset is located.
            val_data: Directory on localhost where
            validation dataset is located.

        """
        config = self.config
        train_dataset = create_dataset(config, train_data, TRAIN)
        val_dataset = (
            create_dataset(config, val_data, VAL) if val_data else None
        )

        label_mappings = train_dataset.label_mappings

        logger.info(f"length of train dataset is {len(train_dataset)}")
        if val_dataset:
            logger.info(f"length of validation dataset is {len(val_dataset)}")

        train_sampler = FasterRCNN.create_sampler(
            is_distributed=self.distributed,
            dataset=train_dataset,
            is_train=True,
        )
        val_sampler = FasterRCNN.create_sampler(
            is_distributed=self.distributed, dataset=val_dataset, is_train=False
        )

        train_loader = dataloader_creator(
            config, train_dataset, train_sampler, TRAIN, self.distributed
        )
        val_loader = dataloader_creator(
            config, val_dataset, val_sampler, VAL, self.distributed
        )
        self.train_loop(
            train_dataloader=train_loader,
            label_mappings=label_mappings,
            val_dataloader=val_loader,
            train_sampler=train_sampler,
        )
        self.writer.close()
        self.kfp_writer.write_metric()

    def train_loop(
        self,
        *,
        train_dataloader,
        label_mappings,
        val_dataloader,
        train_sampler=None,
    ):
        """train on whole range of epochs.

        Args:
            train_dataloader (torch.utils.data.DataLoader):
            label_mappings (dict): a dict of {label_id: label_name} mapping
            val_dataloader (torch.utils.data.DataLoader)
            train_sampler: (torch.utils.data.Sampler)
        """
        model = self.model.to(self.device)
        params = [p for p in model.parameters() if p.requires_grad]
        optimizer, lr_scheduler = FasterRCNN.create_optimizer_lrs(
            self.config, params
        )
        accumulation_steps = self.config.train.get(
            "accumulation_steps", DEFAULT_ACCUMULATION_STEPS
        )
        logger.debug("Start training")
        total_timer = Timer(
            name="total-time", text=const.TIMING_TEXT, logger=logging.info
        )
        total_timer.start()
        for epoch in range(self.config.train.epochs):
            with Timer(
                name=f"epoch-{epoch}-train-time",
                text=const.TIMING_TEXT,
                logger=logging.info,
            ):
                self.train_one_epoch(
                    optimizer=optimizer,
                    data_loader=train_dataloader,
                    epoch=epoch,
                    lr_scheduler=lr_scheduler,
                    accumulation_steps=accumulation_steps,
                )
            if self.distributed:
                train_sampler.set_epoch(epoch)
            self.checkpointer.save(self, epoch=epoch)
            with Timer(
                name=f"epoch-{epoch}-evaluate-time",
                text=const.TIMING_TEXT,
                logger=logging.info,
            ):
                self.evaluate_per_epoch(
                    data_loader=val_dataloader,
                    epoch=epoch,
                    label_mappings=label_mappings,
                )
        total_timer.stop()

    def train_one_epoch(
        self,
        *,
        optimizer,
        data_loader,
        epoch,
        lr_scheduler,
        accumulation_steps,
    ):
        """train per epoch.

        Args:
            optimizer: pytorch optimizer
            data_loader(DataLoader): pytorch dataloader
            epoch (int): lr_scheduler: Pytorch LR scheduler
            lr_scheduler: Pytorch LR scheduler
            accumulation_steps(int): Accumulated Gradients are only updated
            after X steps. This creates an effective batch size of
            batch_size * accumulation_steps
        """
        self.model.train()
        loss_metric = Loss()
        optimizer.zero_grad()
        n_examples = len(data_loader.dataset)
        for i, (images, targets) in enumerate(data_loader):
            images = list(image.to(self.device) for image in images)
            targets = [
                {k: v.to(self.device) for k, v in t.items()} for t in targets
            ]
            loss_dict = self.model(images, targets)

            losses_grad = sum(loss for loss in loss_dict.values())

            loss_dict_reduced = reduce_dict(loss_dict)
            losses = sum(loss for loss in loss_dict_reduced.values())
            loss_metric.update(avg_loss=losses.item(), batch_size=len(targets))

            if not math.isfinite(losses):
                raise BadLoss(
                    f"Loss is {losses}, stopping training. Input was "
                    f"{images, targets}. Loss dict is {loss_dict}"
                )
            elif i % self.config.train.log_frequency == 0:
                intermediate_loss = loss_metric.compute()
                examples_seen = epoch * n_examples + loss_metric.num_examples
                logger.info(
                    f"intermediate loss after mini batch {i} in epoch {epoch} "
                    f"(total training examples: {examples_seen}) is "
                    f"{intermediate_loss}"
                )
                self.writer.add_scalar(
                    "training/intermediate_loss",
                    intermediate_loss,
                    examples_seen,
                )
            losses_grad.backward()
            if (i + 1) % accumulation_steps == 0:
                optimizer.step()
                lr_scheduler.step()
                optimizer.zero_grad()
        self.writer.add_scalar("training/loss", loss_metric.compute(), epoch)
        self.writer.add_scalar(
            "training/running_loss",
            loss_metric.compute(),
            loss_metric.num_examples,
        )
        self.writer.add_scalar(
            "training/lr", optimizer.param_groups[0]["lr"], epoch
        )
        loss_metric.reset()

    def evaluate(self, test_data, **kwargs):
        """evaluate given dataset."""
        config = self.config
        test_dataset = create_dataset(config, test_data, TEST)
        label_mappings = test_dataset.label_mappings
        test_sampler = FasterRCNN.create_sampler(
            is_distributed=self.distributed,
            dataset=test_dataset,
            is_train=False,
        )

        logger.info(f"length of test dataset is {len(test_dataset)}")
        logger.info("Start evaluating estimator: %s", type(self).__name__)

        test_loader = dataloader_creator(
            config, test_dataset, test_sampler, TEST, self.distributed
        )
        self.model.to(self.device)
        self.evaluate_per_epoch(
            data_loader=test_loader,
            epoch=0,
            label_mappings=label_mappings,
            synchronize_metrics=self.sync_metrics,
        )
        self.writer.close()
        self.kfp_writer.write_metric()

    @torch.no_grad()
    def evaluate_per_epoch(
        self,
        *,
        data_loader,
        epoch,
        label_mappings,
        max_detections_per_img=MAX_BOXES_PER_IMAGE,
        synchronize_metrics=True,
    ):
        """Evaluate model performance per epoch.

        Note, torchvision's implementation of faster
        rcnn requires input and gt data for training mode and returns a
        dictionary of losses (which we need to record the loss). We also need to
        get the raw predictions, which is only possible in model.eval() mode,
        to calculate the evaluation metric.
        Args:
            data_loader (DataLoader): pytorch dataloader
            epoch (int): current epoch, used for logging
            label_mappings (dict): a dict of {label_id: label_name} mapping
            max_detections_per_img: max number of targets or predictions allowed
            per example
            is_distributed: whether or not the model is distributed
            synchronize_metrics: whether or not to synchronize evaluation
            metrics across processes
        Returns:

        """
        n_threads = torch.get_num_threads()
        torch.set_num_threads(1)
        self.model.eval()

        loss_metric = Loss()
        for metric in self.metrics.values():
            metric.reset()
        for i, (images, targets) in enumerate(tqdm(data_loader)):
            images = list(image.to(self.device) for image in images)
            targets_raw = [
                {k: v.to(self.device) for k, v in t.items()} for t in targets
            ]
            targets_converted = convert_bboxes2canonical(targets)
            # we need 2 do forward passes, one in eval mode gets us the raw
            # preds and one in train mode gets us the losses
            # pass number 1 to get mAP
            self.model.eval()
            preds_raw = self.model(images)
            converted_preds = convert_bboxes2canonical(preds_raw)
            gt_preds = list(zip(targets_converted, converted_preds))
            if self.distributed and synchronize_metrics:
                all_preds_gt_canonical = gather_gt_preds(
                    gt_preds=gt_preds,
                    device=self.device,
                    max_boxes=max_detections_per_img,
                )
            else:
                all_preds_gt_canonical = gt_preds
            for metric in self.metrics.values():
                metric.update(all_preds_gt_canonical)

            # pass number 2 to get val loss
            self.model.train()
            loss_dict = self.model(images, targets_raw)
            loss_dict_reduced = reduce_dict(loss_dict)
            losses = sum(loss for loss in loss_dict_reduced.values())
            loss_metric.update(avg_loss=losses.item(), batch_size=len(targets))

        self.log_metric_val(label_mappings, epoch)
        val_loss = loss_metric.compute()
        logger.info(f"validation loss is {val_loss}")
        self.writer.add_scalar("val/loss", val_loss, epoch)

        torch.set_num_threads(n_threads)

    def log_metric_val(self, label_mappings, epoch):
        """log metric values.

        Args:
            label_mappings (dict): a dict of {label_id: label_name} mapping
            epoch (int): current epoch, used for logging
        """
        for metric_name, metric in self.metrics.items():
            result = metric.compute()
            logger.debug(result)
            logger.info(f"metric {metric_name} has result: {result}")
            if metric.TYPE == "scalar":
                self.writer.add_scalar(f"val/{metric_name}", result, epoch)
                self.kfp_writer.add_metric(name=metric_name, val=result)
            # TODO (YC) This is hotfix to allow user map between label_id
            # to label_name during model evaluation. In ideal cases this mapping
            # should be available before training/evaluation dataset was loaded.
            # label_id that was missing from label_name should be removed from
            # dataset and the training procedure.
            elif metric.TYPE == "metric_per_label":
                label_results = {
                    label_mappings.get(id, str(id)): value
                    for id, value in result.items()
                }
                self.writer.add_scalars(
                    f"val/{metric_name}-per-class", label_results, epoch
                )
                fig = metric_per_class_plot(metric_name, result, label_mappings)
                self.writer.add_figure(f"{metric_name}-per-class", fig, epoch)

    def save(self, path):
        """Serialize Estimator to path.

        Args:
            path (str): full path to save serialized estimator

        Returns:
            saved full path of the serialized estimator
        """
        save_dict = {
            "model": self.model_without_ddp.state_dict(),
            "config": self.config,
        }
        torch.save(save_dict, path)

        return path

    def load(self, path):
        """Load Estimator from path.

        Args:
            path (str): full path to the serialized estimator
        """
        logger.info(f"loading checkpoint from file")
        checkpoint = torch.load(path, map_location=self.device)
        self.model_without_ddp.load_state_dict(checkpoint["model"])
        loaded_config = copy.deepcopy(checkpoint["config"])
        stored_config = copy.deepcopy(self.config)
        if stored_config != loaded_config:
            logger.debug(
                f"Found difference in estimator config."
                f"Estimator loaded from {path} was trained using config: \n"
                f"{loaded_config}. \nHowever, the current config is: \n"
                f"{self.config}."
            )
        self.model_without_ddp.eval()

    def predict(self, pil_img, box_score_thresh=0.5):
        """Get prediction from one image using loaded model.

        Args:
            pil_img (PIL Image): PIL image from dataset.
            box_score_thresh (float): box score threshold for filter out lower
            score bounding boxes. Defaults to 0.5.

        Returns:
            filtered_pred_annotations (List[BBox2D]): high predicted score
            bboxes from the model.
        """
        img_tensor = transforms.ToTensor()(pil_img).unsqueeze(0)
        img_tensor = img_tensor.to(self.device)
        predicts = self.model_without_ddp(img_tensor)
        predict_annotations = convert_bboxes2canonical(predicts)[0]
        filtered_pred_annotations = [
            box for box in predict_annotations if box.score >= box_score_thresh
        ]
        return filtered_pred_annotations

    @staticmethod
    def create_sampler(is_distributed, *, dataset, is_train):
        """create sample of data.

        Args:
            is_distributed: whether or not the model is distributed
            dataset: dataset obj must have len and __get_item__
            is_train: whether or not the sampler is for training data

        Returns:
            data_sampler: (torch.utils.data.Sampler)

        """
        if is_distributed:
            sampler = torch.utils.data.distributed.DistributedSampler(
                dataset, shuffle=is_train
            )
        elif is_train:
            sampler = torch.utils.data.RandomSampler(dataset)
        else:
            sampler = torch.utils.data.SequentialSampler(dataset)

        return sampler

    @staticmethod
    def create_optimizer_lrs(config, params):
        """create optimizer and learning rate scheduler.

        Args:
            config: (CfgNode): estimator config:
            params: model parameters
        Returns:
            optimizer: pytorch optimizer
            lr_scheduler: pytorch LR scheduler

        """
        if config.optimizer.name == "Adam":
            optimizer = torch.optim.Adam(params, **config.optimizer.args)

            # use fixed learning rate when using Adam
            lr_scheduler = torch.optim.lr_scheduler.LambdaLR(
                optimizer, lambda x: 1.0
            )
        elif config.optimizer.name == "SGD":
            optimizer = torch.optim.SGD(
                params,
                lr=config.optimizer.args.lr,
                momentum=config.optimizer.args.momentum,
                weight_decay=float(config.optimizer.args.weight_decay),
            )
            lr_scheduler = torch.optim.lr_scheduler.StepLR(
                optimizer,
                step_size=config.optimizer.args.lr_step_size,
                gamma=config.optimizer.args.lr_gamma,
            )
        else:
            raise ValueError(
                f"only valid optimizers are 'SGD' and 'Adam' but "
                f"received paramater {config.optimizer.name}"
            )
        return optimizer, lr_scheduler

    @staticmethod
    def get_transform():
        """transform bounding box and tesnor."""
        transforms = [BoxListToTensor(), ToTensor()]
        return Compose(transforms)

    @staticmethod
    def collate_fn(batch):
        """Prepare batch to be format Faster RCNN expects.

        Args:
            batch: mini batch of the form ((x0,x1,x2...xn),(y0,y1,y2...yn))

        Returns:
            mini batch in the form [(x0,y0), (x1,y2), (x2,y2)... (xn,yn)]
        """
        return tuple(zip(*batch))


def create_dataset(config, data_path, split):
    """download dataset from source.

    Args:
        config: (CfgNode): estimator config:
        data_path: Directory on localhost where datasets are located.
        split: train, val, test

    Returns dataset: dataset obj must have len and __get_item__

    """
    dataset = Dataset.create(
        config[split].dataset.name,
        data_path=data_path,
        transforms=FasterRCNN.get_transform(),
        **config[split].dataset.args,
    )
    return dataset


def create_dataloader(
    distributed,
    dataset,
    sampler,
    train,
    *,
    batch_size=1,
    num_workers=0,
    collate_fn=None,
):
    """load dataset and create dataloader.

    Args:
        distributed: wether or not the dataloader is distributed
        dataset: dataset obj must have len and __get_item__
        sampler: (torch.utils.data.Sampler)
        train: whether or not the sampler is for training data
        batch_size: batch_size
        num_workers: num_workers
        collate_fn: Prepare batch to be format Faster RCNN expects
    Returns data_loader:
        torch.utils.data.DataLoader

    """
    if distributed:
        if train:
            batch_sampler = torch.utils.data.BatchSampler(
                sampler, batch_size, drop_last=False
            )

            loader = torch.utils.data.DataLoader(
                dataset,
                batch_sampler=batch_sampler,
                num_workers=num_workers,
                collate_fn=collate_fn,
            )
            return loader
        else:
            loader = torch.utils.data.DataLoader(
                dataset,
                batch_size=batch_size,
                sampler=sampler,
                num_workers=num_workers,
                collate_fn=collate_fn,
                drop_last=False,
            )
            return loader
    else:
        logger.info(f"not creating distributed dataloader")
        data_loader = torch.utils.data.DataLoader(
            dataset,
            batch_size=batch_size,
            sampler=sampler,
            num_workers=num_workers,
            drop_last=train,
            collate_fn=collate_fn,
        )
        return data_loader


def dataloader_creator(config, dataset, sampler, split, distributed):
    """initiate data loading.

    Args:
        config: (CfgNode): estimator config:
        dataset: dataset obj must have len and __get_item__
        sampler: (torch.utils.data.Sampler)
        split: train, val, test
        distributed:
    Returns data_loader:
        torch.utils.data.DataLoader

    """
    is_train = False
    if split == TRAIN:
        is_train = True
    dataloader = create_dataloader(
        distributed=distributed,
        dataset=dataset,
        batch_size=config[split].batch_size,
        sampler=sampler,
        collate_fn=FasterRCNN.collate_fn,
        train=is_train,
    )
    return dataloader


class BadLoss(Exception):
    """pass the exception."""

    pass


class Loss:
    """Record Loss during epoch."""

    def __init__(self):
        """initiate loss."""
        self._sum = 0
        self.num_examples = 0

    def reset(self):
        """reset loss."""
        self._sum = 0
        self.num_examples = 0

    def update(self, avg_loss, batch_size):
        """update loss."""
        self._sum += avg_loss * batch_size
        self.num_examples += batch_size

    def compute(self):
        """compute avg loss.

        Returns (float): avg. loss
        """
        if self.num_examples == 0:
            raise ValueError(
                "Loss must have at least one example "
                "before it can be computed."
            )
        return self._sum / self.num_examples


def convert_bboxes2canonical(bboxes):
    """convert bounding boxes to canonical.

    convert bounding boxes from the format used by pytorch torchvision's
    faster rcnn model into our canonical format, a list of list of BBox2Ds.
    Faster RCNN format:
    https://github.com/pytorch/vision/blob/master/torchvision/models/
    detection/faster_rcnn.py#L45
    Args:
        bboxes (List[Dict[str, torch.Tensor()): A list of dictionaries. Each
        item in the list corresponds to the bounding boxes for one example.
        The dictionary must have the keys 'boxes' and 'labels'. The value for
        'boxes' is (``FloatTensor[N, 4]``): the ground-truth boxes in
        ``[x1, y1, x2, y2]`` format, with values between ``0`` and ``H`` and
        ``0`` and ``W``. The value for labels is (``Int64Tensor[N]``): the
        class label for each ground-truth box. If the dictionary has the key
        `scores` then these values are used for the confidence score of the
        BBox2D, otherwise the score is set to 1.

    Returns (list[List[BBox2D]]):
        Each element in the list corresponds to the
        list of bounding boxes for an example.

    """
    bboxes_batch = []
    for example in bboxes:
        all_coords = example["boxes"]
        num_boxes = all_coords.shape[0]
        labels = example["labels"]
        if "scores" in example.keys():
            scores = example["scores"]
        else:
            scores = torch.FloatTensor([1.0] * num_boxes)
        bboxes_example = []
        for i in range(num_boxes):
            coords = all_coords[i]
            x, y = coords[0].item(), coords[1].item()
            canonical_box = BBox2D(
                x=x,
                y=y,
                w=coords[2].item() - x,
                h=coords[3].item() - y,
                label=labels[i].item(),
                score=scores[i].item(),
            )
            bboxes_example.append(canonical_box)
        bboxes_batch.append(bboxes_example)
    return bboxes_batch


def reduce_dict(input_dict, average=True):
    """Reduce the values in dictionary.

    Reduce the values in the dictionary from all processes so that all processes
    have the averaged results.
    Args:
        input_dict (dict): all the values will be reduced
        average (bool): whether to do average or sum

    Return:
        dict with the same fields as input_dict, after reduction.
    """
    world_size = get_world_size()
    if world_size < 2:
        return input_dict
    with torch.no_grad():
        names = []
        values = []
        # sort the keys so that they are consistent across processes
        for k in sorted(input_dict.keys()):
            names.append(k)
            values.append(input_dict[k])
        values = torch.stack(values, dim=0)
        dist.all_reduce(values)
        if average:
            values /= world_size
        reduced_dict = {k: v for k, v in zip(names, values)}
    return reduced_dict


def prepare_bboxes(bboxes: List[BBox2D]) -> Dict[str, torch.Tensor]:
    """Prepare bounding boxes for model training.

    Args:
        bboxes: mini batch of bounding boxes (not including images).
        Each example is a list of bounding boxes. Torchvision's implementation
        of Faster-RCNN requires bounding boxes to be in the format of a
        dictionary {'labels':[label ids, ...], and
        'boxes': [[xleft, ytop, xright, ybottom of box1],
        [xleft, ytop, xright, ybottom of box2]...]

    Returns:
        bounding boxes in the form that Faster RCNN expects
    """
    labels = []
    box_list = []
    for b in bboxes:
        labels.append(b.label)
        box = [b.x, b.y, b.x + b.w, b.y + b.h]
        if box[0] >= box[2] or box[1] >= box[3]:
            raise ValueError(
                f"box not properly formed with coordinates: "
                f"{box} [xleft, ytop, xright, ybottom]"
            )
        box_list.append(box)

    d = {"boxes": torch.Tensor(box_list), "labels": torch.LongTensor(labels)}
    return d


def canonical2list(bbox: BBox2D):
    """convert a BBox2d into a single list.

    Args:
        bbox:

    Returns:
        attribute list of BBox2D

    """
    return [bbox.label, bbox.score, bbox.x, bbox.y, bbox.w, bbox.h]


def list2canonical(box_list):
    """convert a list into a Bbox2d.

    Args:
        box_list: box represented in list format

    Returns:
        BBox2d

    """
    return BBox2D(
        label=box_list[0],
        score=box_list[1],
        x=box_list[2],
        y=box_list[3],
        w=box_list[4],
        h=box_list[5],
    )


def list3d_2canonical(batch):
    """convert 3d list to canonical.

    convert a list of list of padded targets and predictions per examples where
    bounding boxes are represented by lists into the same format except
    the boxes are represented by the BBox2d class and the padded boxes are
    removed.
    Args:
        batch: [[[gt], [prds]], [[gt], [prds]], ] where gt and prds are list
        of lists

    Returns:
        [([gt],[preds]), ([gt],[preds])... where gt and preds are lists of
        BBox2ds
    """
    all_canonical = []
    for example in batch:
        targets, preds = example
        # todo should break after first nan
        targets_canonnical = [
            list2canonical(target_list)
            for target_list in targets
            if not np.isnan(target_list).any()
        ]
        preds = [
            list2canonical(pred_list)
            for pred_list in preds
            if not np.isnan(pred_list).any()
        ]
        all_canonical.append((targets_canonnical, preds))
    return all_canonical


def pad_box_lists(
    gt_preds: List[Tuple[List[BBox2D], List[BBox2D]]],
    max_boxes_per_img=MAX_BOXES_PER_IMAGE,
):
    """Pad the list of boxes.

    Pad the list of boxes and targets with place holder boxes so that all
    targets and predictions have the same number of elements.
    Args:
        gt_preds (list(tuple(list(BBox2d), (Bbox2d)))): A list of tuples where
        the first element in each tuple is a list of bounding boxes
        corresponding to the targets in an example, and the second element
        in the tuple corresponds to the predictions in that example
        max_boxes_per_img: : maximum number of target boxes and predicted boxes
         per image

    Returns: same format as gt_preds but all examples will have the same
    number of targets and predictions. If there are fewer targets or
    predictions than max_boxes_per_img, then boxes with nan values are added.

    """
    padding_box = BBox2D(
        label=np.nan, score=np.nan, x=np.nan, y=np.nan, w=np.nan, h=np.nan
    )
    for tup in gt_preds:
        target_list, pred_list = tup
        if len(target_list) > max_boxes_per_img:
            raise ValueError(
                f"max boxes per image set to {max_boxes_per_img},"
                f" but there were {len(target_list)} targets"
                f" found."
            )
        if len(pred_list) > max_boxes_per_img:
            raise ValueError(
                f"max boxes per image set to {max_boxes_per_img},"
                f" but there were {len(target_list)} predictions"
                f" found."
            )
        for i in range(max_boxes_per_img - len(target_list)):
            target_list.append(padding_box)
        for i in range(max_boxes_per_img - len(pred_list)):
            pred_list.append(padding_box)
    return gt_preds


def _gt_preds2tensor(gt_preds, max_boxes=MAX_BOXES_PER_IMAGE):
    """convert prediction result to tensor.

    Args:
        gt_preds (list(tuple(list(BBox2d), (Bbox2d)))): A list of tuples where
        the first element in each tuple is a list of bounding boxes
        corresponding to the targets in an example, and the second element
        in the tuple corresponds to the predictions in that example
        max_boxes: maximum number of target boxes and predicted boxes per image

    Returns (Tensor): [[gt, preds], [gt, preds]] where gt and preds are both
    tensors where each element is a 1-d tensor representing a bbox

    """
    gt_preds_padded = pad_box_lists(gt_preds, max_boxes)
    padded_list = []
    for i in range(len(gt_preds_padded)):
        gt_list, target_list = gt_preds_padded[i]
        gt_tensor, target_tensor = (
            [canonical2list(b) for b in gt_list],
            [canonical2list(b) for b in target_list],
        )
        gt_preds_tensor = [gt_tensor, target_tensor]
        padded_list.append(gt_preds_tensor)
    gt_preds_tensor = torch.Tensor(padded_list)
    return gt_preds_tensor


def gather_gt_preds(*, gt_preds, device, max_boxes=MAX_BOXES_PER_IMAGE):
    """gather list of prediction.

    Args:
        gt_preds (list(tuple(list(BBox2d), (Bbox2d)))): A list of tuples where
        the first element in each tuple is a list of bounding boxes
        corresponding to the targets in an example, and the second element
        in the tuple corresponds to the predictions in that example
        device:
        max_boxes: the maximum number of boxes allowed for either targets or
        predictions per image

    Returns (list(tuple(list(BBox2d), (Bbox2d)))): a list in the same format
    as gt_preds but containing all the information across processes e.g. if
    rank 0 has gt_preds = [([box_0], []), ([box_1], [box_2, box_2.5])] and
    rank 1 has gt_preds [([], [box_3]), ([box_4, box_5], [])] then this
    function will return [([box_0], []), ([box_1], [box_2, box_2.5]), ([],
    [box_3]), ([box_4, box_5], [])] the returned list is consistent across all
    processes
    """
    gt_preds_tensor = _gt_preds2tensor(gt_preds, max_boxes)

    all_preds_gt = [
        torch.empty(list(gt_preds_tensor.size())).to(device)
        for i in range(get_world_size())
    ]
    dist.all_gather(tensor_list=all_preds_gt, tensor=gt_preds_tensor.to(device))
    all_preds_gt_canonical = tensorlist2canonical(all_preds_gt)
    return all_preds_gt_canonical


def tensorlist2canonical(tensor_list):
    """convert tensorlist to canonical.

    Converts the gt and predictions into the canonical format and removes
    the boxes with nan values that were added for padding.
    Args:
        tensor_list: [tensor([[gt, prds]), tensor([gt, prds])], ...]

    Returns (list(tuple(list(BBox2d), (Bbox2d)))): A list of tuples where
        the first element in each tuple is a list of bounding boxes
        corresponding to the targets in an example, and the second element
        in the tuple corresponds to the predictions in that example

    """
    gt_preds_canonical = []
    for t in tensor_list:
        gt_preds_canonical += list3d_2canonical(t.tolist())
    return gt_preds_canonical


def metric_per_class_plot(metric_name, data, label_mappings, figsize=(20, 10)):
    """Bar plot for metric per class.

    Args:
        metric_name (str): metric name.
        data (dict): a dictionary of metric per label.
        label_mappings (dict): a dict of {label_id: label_name} mapping
        figsize (tuple): figure size of the plot. Default is (20, 10)

    Returns (matplotlib.pyplot.figure):
        a bar plot for metric per class.
    """
    label_id = [k for k in sorted(data.keys()) if k in label_mappings]
    metric_values = [data[i] for i in label_id]
    label_name = [label_mappings[i] for i in label_id]
    fig = plt.figure(figsize=figsize)
    plt.bar(label_id, metric_values)
    plt.title(f"{metric_name} per class")
    plt.xlabel("label name")
    plt.ylabel(f"{metric_name}")
    plt.xticks(
        label_id, label_name, rotation="vertical",
    )
    plt.margins(0.2)
    plt.subplots_adjust(bottom=0.3)
    return fig


class BoxListToTensor:
    """transform to bboxes to Tensor."""

    def __call__(self, image, target):
        """transform target to bboxes."""
        target = prepare_bboxes(target)
        return image, target


class ToTensor:
    """transform to tesnor."""

    def __call__(self, image, target):
        """transform image to tesnor."""
        image = torchvision.transforms.functional.to_tensor(image)
        return image, target



import logging
import os
import shutil
import tempfile
import zipfile

import numpy as np
from PIL import Image
from pyquaternion import Quaternion

import datasetinsights.constants as const
from datasetinsights.io.bbox import BBox3d
from datasetinsights.io.gcs import GCSClient

from .base import Dataset
from .nuscenes import Box

logger = logging.getLogger(__name__)
KITTI_GCS_PATH = "data/kitti"
SPLITS = ["train", "test", "val", "trainval"]  # test refers to KITTI's test
# set which doesn't have labels

KITTI = "kitti"
NUSCENES = "nuscenes"
VALID_FORMATS = [KITTI, NUSCENES]
SAMPLEX_INDICES_FILE = "samples.txt"
ZIP_FILES = [
    "data_object_calib.zip",
    "data_object_image_2.zip",
    "data_object_label_2.zip",
]


class KittiBox3d:
    """
    class to represent a bounding box for the kitti dataset. Note that this
    style of bounding box is not primarily supported. The canonical 3d bounding
    box class for this repo is the class BBox3D.
    Reference code for KittiBox3d found at
    http://www.cvlibs.net/datasets/kitti/eval_object.php?obj_benchmark=3d
    https://github.com/bostondiditeam/kitti/tree/master/resources/devkit_object
    To convert from kitti style to nuscenes use the method convert_kitti2nu.
    """

    def __init__(
        self,
        *,
        label: str,
        position,
        dimensions,
        angle,
        sample_idx,
        score: float = 1.0,
    ):
        """

        Args:
            label: class label possibilities are 'car' 'pedestrian' and
            'cyclist'
            position: x,y,z in camera coordinates (meters)
            dimensions: length, height, width  (meters)
            angle: angle about vertical axis in radians range [-pi, pi]
            sample_idx: the index corresponding to the sample data (image), the
             image is stored in
            image_2/{sample_index:06d}.png
            score: confidence score (defaults to zero, to be used for ground
            truth)
        """
        self.label = label
        self.position = position
        self.dimensions = dimensions
        self.angle = angle
        self.score = score
        self.sample_idx = sample_idx


# todo add tests
class KittiTransforms:
    """
    Class to hold transformation matrices for a kitti data sample see more at
    https://github.com/yanii/kitti-pcl/blob/master/KITTI_README.TXT
    """

    def __init__(self, calib_filename):
        self.lines = [line.rstrip() for line in open(calib_filename)]

    def _get_velo2camera(self):
        """
        matrix takes a point in Velodyne coordinates and transforms it into the
        coordinate system of the left video camera. Likewise it serves as a
        representation of the Velodyne coordinate frame in camera coordinates.
        Returns: Combined translation and rotation matrix

        """
        velo_to_cam = np.array(
            self.lines[5].strip().split(" ")[1:], dtype=np.float32
        )
        velo_to_cam.resize((3, 4))
        return velo_to_cam

    @property
    def velo_to_cam_rotation(self):
        """
        Rotation matrix takes a point in Velodyne coordinates and transforms it
         into the
        coordinate system of the left video camera. Likewise it serves as a
        representation of the Velodyne coordinate frame in camera coordinates.
        Returns: Rotation matrix

        """
        velo_to_cam = self._get_velo2camera()
        return velo_to_cam[:, :3]

    @property
    def velo_to_cam_translation(self):
        """
        Translation matrix takes a point in Velodyne coordinates and transforms
         it into the
        coordinate system of the left video camera. Likewise it serves as a
        representation of the Velodyne coordinate frame in camera coordinates.
        Returns: Translation matrix

        """
        velo_to_cam = self._get_velo2camera()
        return velo_to_cam[:, 3]

    @property
    def r0_rect(self) -> np.ndarray:
        """

        Returns: Quaternion to rectify camera frame.

        """
        r0_rect = np.array(
            self.lines[4].strip().split(" ")[1:], dtype=np.float32
        )
        r0_rect.resize((3, 3))
        return r0_rect

    @property
    def projection_mat_left(self):
        """

        Returns: Projection matrix for left image (to project bounding box
        coordinate to image coordinates).

        """
        p_left = np.array(
            self.lines[2].strip().split(" ")[1:], dtype=np.float32
        )
        return p_left.resize((3, 4))

    @property
    def projection_rect_combined(self):
        """
        Merge rectification and projection into one matrix.

        Returns: combined rectification and projection matrix

        """
        p_combined = np.eye(4)
        p_combined[:3, :3] = self.r0_rect
        p_combined = np.dot(self.projection_mat_left, p_combined)
        return p_combined


def convert_kitti2nu(
    *, bbox: KittiBox3d, transforms: KittiTransforms
) -> BBox3d:
    """
    convert a bounding box from kitti format to nuscenes format

    Args:
        bbox: bounding box in kitti format
        transforms: camera transforms

    Returns:

    """
    center = bbox.position
    wlh = [
        bbox.dimensions[2],
        bbox.dimensions[0],
        bbox.dimensions[1],
    ]  # lhw -> wlh bbox['wlh']
    yaw_camera = bbox.angle
    name = bbox.label
    score = bbox.score

    # The Box class coord system is oriented the same way as as KITTI LIDAR: x
    # forward, y left, z up.
    # For rotation confer: http://www.cvlibs.net/datasets/kitti/setup.php.

    # 1: Create box in Box coordinate system with center at origin.
    # The second rotation in yaw_box transforms the coordinate frame from the
    # object frame
    # to KITTI camera frame. The equivalent cannot be naively done afterwards,
    # as it's box rotation
    # around the local object coordinate frame, rather than the camera frame.
    quat_box = Quaternion(axis=(0, 1, 0), angle=yaw_camera) * Quaternion(
        axis=(1, 0, 0), angle=np.pi / 2
    )
    box = Box([0.0, 0.0, 0.0], wlh, quat_box, name=name)

    # 2: Translate: KITTI defines the box center as the bottom center of the
    # vehicle. We use true center,
    # so we need to add half height in negative y direction, (since y points
    # downwards), to adjust. The
    # center is already given in camera coord system.
    box.translate(center + np.array([0, -wlh[2] / 2, 0]))

    # 3: Transform to KITTI LIDAR coord system. First transform from rectified
    # camera to camera, then
    # camera to KITTI lidar.
    box.rotate(Quaternion(matrix=transforms.r0_rect).inverse)
    box.translate(-transforms.velo_to_cam_translation)
    box.rotate(Quaternion(matrix=transforms.velo_to_cam_rotation).inverse)

    # Set score or NaN.
    box.score = score

    # Set dummy velocity.
    box.velocity = np.array((0.0, 0.0, 0.0))
    box = BBox3d(
        translation=box.center,
        size=box.wlh,
        rotation=box.orientation,
        label=box.name,
        score=box.score,
        velocity=box.velocity,
        sample_token=bbox.sample_idx,
    )
    return box


def read_kitti_calib(filename):
    """Read the camera 2 calibration matrix from box text file"""

    with open(filename) as f:
        for line in f:
            data = line.split(" ")
            if data[0] == "P2:":
                calib = np.array([float(x) for x in data[1:13]])
                return calib.reshape((3, 4))

    raise FileNotFoundError(
        "Could not find entry for P2 in calib file {}".format(filename)
    )


def read_kitti_objects(filename):
    objects = list()
    with open(filename, "r") as fp:

        # Each line represents box single object
        for line in fp:
            objdata = line.split(" ")
            if not (14 <= len(objdata) <= 15):
                raise IOError("Invalid KITTI object file {}".format(filename))

            # Parse object data
            objects.append(
                KittiBox3d(
                    label=objdata[0],
                    dimensions=[
                        float(objdata[10]),
                        float(objdata[8]),
                        float(objdata[9]),
                    ],
                    position=[float(p) for p in objdata[11:14]],
                    angle=float(objdata[14]),
                    score=float(objdata[15]) if len(objdata) == 16 else 1.0,
                    sample_idx=os.path.basename(filename),
                )
            )
    return objects


class Kitti(Dataset):
    """
    dataloader for kitti dataset
    http://www.cvlibs.net/datasets/kitti/eval_object.php?obj_benchmark=3d
    """

    def __init__(
        self,
        root=os.path.join(const.DEFAULT_DATA_ROOT, "kitti"),
        split="train",
        indices_file: str = None,
    ):
        """

        Args:
            root: path to where data already exists or where it will be
            downloaded to
            split: which split of the data to use. Can be: 'train', 'test',
            'val', 'trainval'. Can either specify split
            or indices file but not both.
            indices_file: file containing indices to use. Can either specify
            split or indices file but not both.
        """
        if split not in SPLITS:
            raise ValueError(
                f"invlaid value for split: {split},"
                f" possible values are: {SPLITS}"
            )
        if split is None and indices_file is None:
            raise ValueError(
                f"Cannot specify both indices file and split, must choose "
                f"one."
            )
        # todo should probably have separate val set with labels
        kitti_split = "testing" if split == "test" else "training"
        # self.root = os.path.join(root, kitti_split)
        self.root = root
        if not os.path.exists(self.root):
            os.makedirs(self.root)
        self.split = split
        downloaded = self._check_downloaded()
        if not downloaded:
            logger.info(f"no local copy of dataset found.")
            self.download(cloud_path=KITTI_GCS_PATH)
        else:
            logger.info("local copy of dataset found, will not download")
        indices_file = indices_file or os.path.join(
            self.root, SAMPLEX_INDICES_FILE
        )
        self.indices = self._read_indices_file(filename=indices_file)
        self.root = os.path.join(root, kitti_split)

    def _check_downloaded(self):
        for z in ZIP_FILES:
            p = os.path.join(self.root, z)
            if not os.path.exists(p):
                logger.info(f"could not find file {p}")
                return False
        return True

    def __len__(self):
        return len(self.indices)

    def _get_calib_filename(self, idx):
        calib_file = os.path.join(self.root, f"calib/{idx:06d}.txt")
        return calib_file

    def _get_label_filename(self, idx):
        label_file = os.path.join(self.root, f"label_2/{idx:06d}.txt")
        return label_file

    def __getitem__(self, index):
        idx = self.indices[index]

        # Load image
        img_file = os.path.join(self.root, f"image_2/{idx:06d}.png")
        image = Image.open(img_file)

        # Load calibration matrix
        calib_file = os.path.join(self.root, f"calib/{idx:06d}.txt")
        calib = read_kitti_calib(calib_file)
        nu_transform = KittiTransforms(calib_filename=calib_file)

        # Load annotations
        label_file = os.path.join(self.root, f"label_2/{idx:06d}.txt")
        objects = read_kitti_objects(label_file)
        bboxes = [
            convert_kitti2nu(bbox=o, transforms=nu_transform) for o in objects
        ]
        return idx, image, calib, bboxes

    def _read_indices_file(self, filename):
        """

        Args:
            filename: path to file which contains kitti sample indices_file

        Returns: list of indices_file

        """
        with open(filename) as f:
            return [int(val) for val in f]

    def _download_sample_indices_file(
        self, *, cloud_client, object_key=None, local_file=None
    ):
        local_file = local_file or os.path.join(self.root, SAMPLEX_INDICES_FILE)
        object_key = object_key or f"{KITTI_GCS_PATH}/splits/{self.split}.txt"
        cloud_client.download(
            local_path=self.root, bucket=const.GCS_BUCKET, key=object_key
        )
        return local_file

    def download_kitti_zips(self, cloud_client, cloud_path=KITTI_GCS_PATH):
        calib_zip_key = f"{cloud_path}/data_object_calib.zip"
        left_images_zip_key = f"{cloud_path}/data_object_image_2.zip"
        left_image_labels_zip_key = f"{cloud_path}/data_object_label_2.zip"
        all_zips = [
            calib_zip_key,
            left_images_zip_key,
            left_image_labels_zip_key,
        ]
        local_zips = []
        for z in all_zips:
            local_path = os.path.join(self.root, z.split("/")[-1])
            cloud_client.download(
                local_path=self.root, bucket=const.GCS_BUCKET, key=z
            )
            local_zips.append(local_path)

        calib_zip, local_left_images_zip, local_labels_zip = [
            os.path.abspath(z) for z in local_zips
        ]
        return calib_zip, local_left_images_zip, local_labels_zip

    def _unzip2dir(self, *, zip_path, src, dst):
        """

        Args:
            zip_path: path to zip file
            src: the path within the unziped files to the file (or dir) to move
            dst: where to move the file (or dir) specified in src to


        """
        logger.info(f"extracting from {src} to {dst} ")
        with tempfile.TemporaryDirectory() as tmp:
            with zipfile.ZipFile(zip_path, "r") as zip_dir:
                zip_dir.extractall(tmp)
                shutil.move(os.path.join(tmp, src), dst)

    def download(self, cloud_path=KITTI_GCS_PATH):
        logger.info(f"downloading kitti dataset from cloud storage")
        # todo is currently only downloading left color images
        cloud_client = GCSClient()
        self._download_sample_indices_file(cloud_client=cloud_client)
        calib_zip, left_images_zip, labels_zip = self.download_kitti_zips(
            cloud_client=cloud_client
        )
        with zipfile.ZipFile(left_images_zip, "r") as zip_ref:
            zip_ref.extractall(self.root)
        testing_dir = os.path.join(self.root, "testing")
        training_dir = os.path.join(self.root, "training")
        self._unzip2dir(
            zip_path=calib_zip,
            src=os.path.join("testing", "calib"),
            dst=os.path.join(testing_dir, "calib"),
        )
        self._unzip2dir(
            zip_path=calib_zip,
            src=os.path.join("training", "calib"),
            dst=os.path.join(training_dir, "calib"),
        )
        self._unzip2dir(
            zip_path=labels_zip,
            src=os.path.join("training", "label_2"),
            dst=os.path.join(training_dir, "label_2"),
        )



import logging
import os
import tarfile
import tempfile
from typing import Dict, List

from nuscenes.eval.detection import loaders
from nuscenes.nuscenes import NuScenes
from nuscenes.utils.data_classes import Box
from torch.utils.data import Dataset

import datasetinsights.constants as const
from datasetinsights.io.bbox import BBox3d
from datasetinsights.io.gcs import GCSClient

logger = logging.getLogger(__name__)

NUSCENES_GCS_PATH = "data/nuscenes"
NUSCENES_LOCAL_PATH = "sets/nuscenes"
CAMERA_KEYS = [
    "CAM_FRONT",
    "CAM_FRONT_RIGHT",
    "CAM_BACK_RIGHT",
    "CAM_BACK",
    "CAM_BACK_LEFT",
    "CAM_FRONT_LEFT",
]
RADAR_KEYS = [
    "RADAR_FRONT",
    "RADAR_FRONT_LEFT",
    "RADAR_FRONT_RIGHT",
    "RADAR_BACK_LEFT",
    "RADAR_BACK_RIGHT",
]
LIDAR_KEY = "LIDAR_TOP"
SENSOR_DATA_KEYS = CAMERA_KEYS + RADAR_KEYS + [LIDAR_KEY]
GLOBAL = "global"
SENSOR = "sensor"
COORDINATES = [GLOBAL, SENSOR]
NUM_TRAINVAL_TARBALLS = 10
VERSIONS = ["v1.0"]
SPLITS = ["train", "val"]


# TODO: This should be a child class of abstract Dataset class
class NuscenesDataLoader(Dataset):
    """ Nuscenes Dataset https://www.nuscenes.org/
    Args:
        data_root (str): root directory prefix of all datasets
        split: test or valuation. see SPLITS constant for possible values
        version: numerical version of nuscenes dataset.
        sensor_data: what sensor data to use, can be any of the strings in
        SENSOR_DATA_KEYS
        full_dataset: whether to use the full dataset. If false, use the mini
        dataset
        download: whether or no to download the nuscenes dataset from gcs. If
        false, it is assumed the
        dataset is stored at the location specified by root
        coordinates: whether to use coordinates in the local coordinate system
        of the sensor or the global
        log coordinates. Options: 'sensor' or 'global'

    Attributes:
        root: directory where nuscenes data is stored. If no value is provided,
            the specified version of nuscenes is downloaded.
    """

    def __init__(
        self,
        *,
        data_root=const.DEFAULT_DATA_ROOT,
        split="train",
        version="v1.0",
        full_dataset: bool = False,
        download=False,
        sensor_data: List[str] = SENSOR_DATA_KEYS,
        coordinates="global",
    ):
        self.root = os.path.join(data_root, NUSCENES_LOCAL_PATH)
        if version not in VERSIONS:
            raise ValueError(
                f"version provided was {version} but only valid versions are: "
                f"{VERSIONS}"
            )
        if full_dataset is False:
            split = f"mini_{split}"
            version = f"{version}-mini"
        else:
            version = f"{version}-trainval"
        if download:
            self.download(version=version)
        nu = NuScenes(dataroot=self.root, version=version)
        self.nu = nu
        self.scenes = nu.scene
        self.split = split
        self.sample_tokens = loaders.load_gt(
            nusc=nu, eval_split=split
        ).sample_tokens
        for s in sensor_data:
            if s not in SENSOR_DATA_KEYS:
                raise ValueError(
                    f"sensor key: {s} is not a valid sensor. "
                    f"Valid sensors are: {SENSOR_DATA_KEYS}"
                )
        self.data_keys = sensor_data
        if coordinates not in COORDINATES:
            raise ValueError(
                f"coordinates can only be one of {COORDINATES} "
                f"but {self.coordinates} was given."
            )
        self.coordinates = coordinates

    def _merge_data_blobs(self, blob, canonical_dir):
        """Merge the data blobs together which comprise the v1.0-trainval
        dataset.
        The structure of each data blob in the nuscenes dataset is
        -v1.0-trainval01_blobs
         |-samples
           |-CAM_FRONT
           |-...all sensors
         |-sweeps
           |-CAM_FRONT
           |-...all sensors
        we need to merge all the contents into one single canonical directory
         which will have the format
        -canonical_dir
         |-samples
           |-CAM_FRONT
           |-...all sensors
         |-sweeps
           |-CAM_FRONT
           |-...all sensors
        Args:
            blob: filename of data blob
            canonical_dir: filename of directory to store all samples find in
            all datablobs
        """
        for d in ["samples", "sweeps"]:
            for sub_dir in SENSOR_DATA_KEYS:
                data_type_dir = os.path.join(blob, d, sub_dir)
                for name in os.listdir(data_type_dir):
                    src = os.path.join(data_type_dir, name)
                    dest = os.path.join(canonical_dir, d, sub_dir, name)
                    os.rename(src, dest)

    def _download_trainval_tars(self) -> (str, List[str]):
        """

        Returns:
            Tuple: First item is the path to the meta data dir, and the second
             item is a list of paths to the
            tarballs comprising the nuscenes trainval dataset
        """
        data_tarballs = [
            f"{NUSCENES_GCS_PATH}/v1.0-trainval0{1 + i}_blobs.tar"
            for i in range(NUM_TRAINVAL_TARBALLS - 1)
        ]
        data_tarballs.append(f"{NUSCENES_GCS_PATH}/v1.0-trainval10_blobs.tar")
        trainval_tars = []
        for data in data_tarballs:
            local_dest = os.path.join(self.root, os.path.basename(data))
            trainval_tars.append(local_dest)
            self.cloud_client.download(
                local_path=self.root, bucket=const.GCS_BUCKET, key=data
            )
        meta_gcs_key = f"{NUSCENES_GCS_PATH}/v1.0-trainval_meta.tar"
        meta_local_path = os.path.join(self.root, "v1.0-trainval_meta.tar")
        self.cloud_client.download(
            local_path=self.root, bucket=const.GCS_BUCKET, key=meta_gcs_key,
        )
        return meta_local_path, trainval_tars

    def _download_mini(self):
        mini_tar_path = f"{NUSCENES_GCS_PATH}/v1.0-mini.tar"
        local_tar_path = f"{self.root}/v1.0-mini.tar"
        logger.info("downloading mini dataset from gcs")
        self.cloud_client.download(
            local_path=self.root, bucket=const.GCS_BUCKET, key=mini_tar_path,
        )
        with tarfile.open(local_tar_path) as t:
            t.extractall(self.root)
        return self.root

    def download(self, version="v1.0-mini"):
        """
        download the nuscenes dataset version specified to /data/sets/nuscenes
        """
        self.cloud_client = GCSClient()
        if version == "v1.0-mini":
            return self._download_mini()
        elif version == "v1.0-trainval":
            meta_local_path, trainval_tars = self._download_trainval_tars()
            with tarfile.open(meta_local_path) as t:
                t.extractall(self.root)
            first_data_blob = trainval_tars[0]
            with tarfile.open(first_data_blob) as t:
                logger.info(f"extracting files from {first_data_blob}")
                t.extractall(self.root)
            os.remove(trainval_tars[0])
            for tar_path in trainval_tars[1:]:
                with tarfile.open(tar_path) as t:
                    logger.info(f"extracting files from {tar_path}")
                    with tempfile.TemporaryDirectory() as tmpdirname:
                        t.extractall(tmpdirname)
                        self._merge_data_blobs(
                            blob=tmpdirname, canonical_dir=self.root
                        )
            return self.root
        else:
            raise ValueError()

    def _nu_box2bbox3d(self, nu_box: Box, idx) -> BBox3d:
        width, length, height = nu_box.wlh
        return BBox3d(
            rotation=nu_box.orientation,
            translation=nu_box.center,
            label=nu_box.label,
            score=nu_box.score,
            velocity=nu_box.velocity,
            size=[width, height, length],
            sample_token=idx,
        )

    def __getitem__(self, item) -> Dict[str, List[BBox3d]]:
        """
        Args:
            item: index
        Returns:
            Dictionary: mapping the name of each sensor to the list of bounding
             boxes found in that sensor's data

        """
        data = {}
        sample_token = self.sample_tokens[item]
        sample_timeframe = self.nu.get("sample", sample_token)
        if self.coordinates == SENSOR:
            for sensor in self.data_keys:
                sensor_token = sample_timeframe["data"][sensor]
                data_path, nu_boxes, camera = self.nu.get_sample_data(
                    sensor_token
                )
                bboxes = [self._nu_box2bbox3d(b, item) for b in nu_boxes]
                data[sensor] = bboxes
        elif self.coordinates == GLOBAL:
            for sensor in self.data_keys:
                sensor_token = sample_timeframe["data"][sensor]
                nu_boxes = self.nu.get_boxes(sensor_token)
                bboxes = [self._nu_box2bbox3d(b, item) for b in nu_boxes]
                data[sensor] = bboxes
        else:
            raise ValueError(
                f"coordinates can only be one of {COORDINATES} but "
                f"{self.coordinates} was given."
            )
        return data

    def __len__(self):
        return len(self.sample_tokens)



import glob
import json
import logging
import os
import shutil
import zipfile
from pathlib import Path

import tensorflow as tf
from google.protobuf import text_format
from PIL import Image

import datasetinsights.constants as const
from datasetinsights.io.bbox import BBox2D
from datasetinsights.io.gcs import GCSClient

from .base import Dataset
from .exceptions import DatasetNotFoundError
from .protos import string_int_label_map_pb2

logger = logging.getLogger(__name__)


class GroceriesReal(Dataset):
    """Unity's Groceries Real Dataset.

    During the class instantiation, it would check whether the data is
    downloaded or not. If there is no dataset, it would raise an error.
    Please make sure you download the dataset before use this class.

    The dataset contain the following structures:

    * images/

        * IMG_4185.JPG
        * IMG_4159.JPG
        * ...
    * `annotation_definition.json` file stores label_id to label_name mappings
    * `annotations.json` file stores 2d bounding box records:

    .. code-block:: json
        [
            {
                "file_name": "IMG_4185.JPG",
                "annotations":[
                    {"label_id": 11, "bboxs":[1692, 1386, 768, 133}
                ]
            },
            ...
        ]
    * `groceries_real_<split>.txt` file which stores space deplimted list
    of the indices in the annotations.json file, for a given data split.


    Attributes:
        dataset_directory (str): Local directory where groceries_real dataset
            is saved.
        split (str): Indicate split type of the dataset. Allowed value are
            (train|val|test|test_high_ratio|test_low_ratio)
            test_high_ratio: Split of test dataset with high
            foreground-to-background ratio.
            test_low_ratio: Split of test dataset with low
            foreground-to-background ratio.
        transforms: Callable transformation
        annotations (list): A list of dict that contain image file path
            and 2D bounding box annotations.
        split_indices (list): A list of indices for this dataset split.
        label_mappings (dict): a dict of {label_id: label_name} mapping
        version (str): version of GroceriesReal dataset, e.g. "v3".
        default version="v3".
    """

    SPLITS = {
        "train": "groceries_real_train.txt",
        "val": "groceries_real_val.txt",
        "test": "groceries_real_test.txt",
        # Split train dataset into small/medium (10%/50%) subset
        "train_small": "groceries_real_train_small.txt",
        "train_medium": "groceries_real_train_medium.txt",
        # Split test dataset into high/low foreground to background ratio
        "test_high_ratio": "groceries_real_test_high_ratio.txt",
        "test_low_ratio": "groceries_real_test_low_ratio.txt",
        # Split test dataset into high/low contrast. Images in high contrast
        # split tend to have more complicated shadow patterns
        "test_low_contrast": "groceries_real_test_low_contrast.txt",
        "test_high_contrast": "groceries_real_test_high_contrast.txt",
    }

    ANNOTATION_FILE = "annotations.json"
    DEFINITION_FILE = "annotation_definitions.json"
    ARCHIVE_FILES = {"v3": "v3.zip"}
    SUBFOLDER = "groceries"

    def __init__(
        self,
        *,
        data_path=const.DEFAULT_DATA_ROOT,
        split="train",
        transforms=None,
        version="v3",
        **kwargs,
    ):
        """
        Args:
            data_path (str): Directory on localhost where datasets are located.
            split (str): Indicate split type of the dataset.
            transforms: callable transformation
            version (str): version of GroceriesReal dataset
        """
        self._data_path = self._preprocess_dataset(data_path, version)

        valid_splits = tuple(self.SPLITS.keys())
        if split not in valid_splits:
            raise ValueError(
                f"Invalid value for split: {split}. Allowed values "
                f"are: {valid_splits}"
            )
        self.split = split

        logger.info(
            f"Using split {split} and version {version} for groceries real "
            "dataset."
        )

        self.version = version
        self.transforms = transforms

        self.annotations = self._load_annotations()
        self.split_indices = self._load_split_indices()
        self.label_mappings = self._load_label_mappings()

    def __getitem__(self, idx):
        """
        Args:
            idx (int): index of the dataset catalog

        Returns:
            tuple: A pair of PIL image and list of BBox2D object.
        """
        data_idx = self.split_indices[idx]
        image_data = self.annotations[data_idx]
        image_filename = image_data["file_name"]
        annotations = image_data["annotations"]
        image_filename = self._filepath(os.path.join("images", image_filename))
        image = Image.open(image_filename)
        bboxes = [self._convert_to_bbox2d(ann) for ann in annotations]
        if self.transforms:
            image, bboxes = self.transforms(image, bboxes)

        return image, bboxes

    def __len__(self):
        return len(self.split_indices)

    def _filepath(self, filename):
        """Local file path relative to data_path
        """
        return os.path.join(self._data_path, filename)

    def _load_annotations(self):
        """Load annotation from annotations.json file

        Returns:
            list: A list of annotations stored in annotations.json file
        """
        annotation_file = self._filepath(self.ANNOTATION_FILE)
        with open(annotation_file) as f:
            json_data = json.load(f)

        return json_data

    def _load_split_indices(self):
        """Load the data indices txt file.

        Returns:
            list: A list of data indices in annotations.json file.
        """
        split_file = self.SPLITS.get(self.split)
        indices_file = self._filepath(split_file)

        with open(indices_file) as txt_file:
            idx_data = [int(i) for i in txt_file.readline().split()]

        return idx_data

    def _load_label_mappings(self):
        """Load label mappings.

        Returns:
            dict: A dict containing {label_id: label_name} mappings.
        """
        definition_file = self._filepath(self.DEFINITION_FILE)
        with open(definition_file) as json_file:
            init_definition = json.load(json_file)["annotation_definitions"][0]
            label_mappings = {
                m["label_id"]: m["label_name"] for m in init_definition["spec"]
            }

        return label_mappings

    @staticmethod
    def _convert_to_bbox2d(single_bbox):
        """Convert the bbox record to BBox2D objects.

        Args:
            single_bbox (dict): raw bounding box information

        Return:
            canonical_bbox (BBox2D): canonical bounding box
        """
        label = single_bbox["label_id"]
        bbox = single_bbox["bbox"]

        canonical_bbox = BBox2D(
            x=bbox[0], y=bbox[1], w=bbox[2], h=bbox[3], label=label
        )
        return canonical_bbox

    @staticmethod
    def _preprocess_dataset(data_path, version):
        """ Preprocess dataset inside data_path and un-archive if necessary.

            Args:
                data_path (str): Path where dataset is stored.
                version (str): groceries_real dataset version.

            Return:
                Path of the dataset files.
            """
        if GroceriesReal.is_dataset_files_present(data_path):
            return data_path

        archive_file = Path(data_path) / GroceriesReal.ARCHIVE_FILES[version]
        if not archive_file.exists():
            raise DatasetNotFoundError(
                f"Expecting a file {archive_file} under {data_path}"
            )

        unarchived_path = Path(data_path) / GroceriesReal.SUBFOLDER
        shutil.unpack_archive(
            filename=archive_file, extract_dir=unarchived_path
        )

        return unarchived_path

    @staticmethod
    def is_dataset_files_present(data_path):
        return (
            os.path.isdir(data_path)
            and any(glob.glob(f"{data_path}/*.json"))
            and any(glob.glob(f"{data_path}/*.txt"))
        )


class GoogleGroceriesReal(Dataset):
    """Google GroceriesReal Dataset

    This dataset include groceries images and annotations in this paper_.
    Please reach out to the authors to request the dataset.

    The dataset should contain the following structures:

    * Train Data/
        file0.tfexample
        file1.tfexample
        ...
    * Eval Data/
        file234.tfexample
        ...
    * Label Map/label_map_with_path_64_retail.txt

    .. _paper: https://arxiv.org/abs/1902.09967

    Attributes:
        root (str): Local directory of the dataset is.
        split (str): Indicate split type of the dataset.
            Allowed value are (train|test).
        transforms: Callable transformation
        label_mappings (dict): a dict of {label_id: label_name} mapping
            loaded from label_map_with_path_64_retail.txt
    """

    GCS_PATH = "data/google_groceries"
    LOCAL_PATH = "google_groceries"
    SPLITS_ZIP = {
        "train": "Train Data.zip",
        # Their dataset does not provide separate validateion/test dataset.
        # Using "eval" as test dataset.
        "test": "Eval Data.zip",
    }
    SPLITS_FOLDER = {
        "train": "Train Data",
        "test": "Eval Data",
    }
    LABEL_ZIP = "Label Map.zip"
    LABEL_FILE = "Label Map/label_map_with_path_64_retail.txt"
    TF_FEATURES = {
        "image/channels": tf.io.FixedLenFeature([], tf.int64),
        "image/encoded": tf.io.FixedLenFeature([], tf.string),
        "image/format": tf.io.FixedLenFeature([], tf.string),
        "image/width": tf.io.FixedLenFeature([], tf.int64),
        "image/height": tf.io.FixedLenFeature([], tf.int64),
        "image/object/bbox/occluded": tf.io.FixedLenFeature([], tf.float32),
        "image/object/bbox/xmax": tf.io.VarLenFeature(tf.float32),
        "image/object/bbox/xmin": tf.io.VarLenFeature(tf.float32),
        "image/object/bbox/ymax": tf.io.VarLenFeature(tf.float32),
        "image/object/bbox/ymin": tf.io.VarLenFeature(tf.float32),
        "image/object/class/label": tf.io.VarLenFeature(tf.int64),
        "image/object/class/text": tf.io.VarLenFeature(tf.string),
    }

    def __init__(self, data_root, split="train", transforms=None):
        """Initialize GoogleGroceriesReal Dataset

        Args:
            data_root (str): Root directory prefix of datasets.
            split (str): Indicate split type of the dataset.
            transforms: callable transformations.
        """
        valid_splits = tuple(self.SPLITS_ZIP.keys())
        if split not in valid_splits:
            raise ValueError(
                f"Invalid value for split: {split}. Allowed values "
                f"are: {valid_splits}"
            )
        self.split = split
        self.root = os.path.join(data_root, self.LOCAL_PATH)
        self.download()

        self._tfexample_files = self._glob_tfexamples()
        self.transforms = transforms
        self.label_mappings = self._load_label_mappings()

    def __getitem__(self, index):
        """
        Args:
            index (int): index of the dataset catalog

        Returns:
            tuple: A pair of PIL image and list of BBox2D objects.
        """
        tfexample_file = self._tfexample_files[index]
        with open(tfexample_file, "rb") as f:
            record = f.read()
        raw_record = tf.io.parse_single_example(record, self.TF_FEATURES)
        image = self._load_image(raw_record)
        bboxes = self._load_bounding_boxes(raw_record)

        if self.transforms:
            image, bboxes = self.transforms(image, bboxes)

        return image, bboxes

    def __len__(self):
        return len(self._tfexample_files)

    def download(self):
        """Download dataset from GCS
        """
        cloud_path = f"gs://{const.GCS_BUCKET}/{self.GCS_PATH}"
        client = GCSClient()
        # download label file
        label_zip = self.LABEL_ZIP
        client.download(url=cloud_path, local_path=self.root)
        with zipfile.ZipFile(label_zip, "r") as zip_dir:
            zip_dir.extractall(self.root)

        # download tfexamples for a dataset split
        tfexamples_zip = self.SPLITS_ZIP.get(self.split)
        client.download(url=cloud_path, local_path=self.root)
        with zipfile.ZipFile(tfexamples_zip, "r") as zip_dir:
            zip_dir.extractall(self.root)

    def _glob_tfexamples(self):
        split_folder = self.SPLITS_FOLDER.get(self.split)
        path = Path(self.root) / split_folder
        filenames = [str(p) for p in path.glob("*.tfexample")]

        return filenames

    def _load_image(self, raw_record):
        tf_img = raw_record["image/encoded"]
        np_img = tf.io.decode_image(tf_img).numpy()
        pil_img = Image.fromarray(np_img)

        return pil_img

    def _load_bounding_boxes(self, raw_record):
        img_width = raw_record["image/width"].numpy()
        img_height = raw_record["image/height"].numpy()
        label = tf.sparse.to_dense(
            raw_record["image/object/class/label"]
        ).numpy()
        xmin = (
            tf.sparse.to_dense(raw_record["image/object/bbox/xmin"]).numpy()
            * img_width
        )
        xmax = (
            tf.sparse.to_dense(raw_record["image/object/bbox/xmax"]).numpy()
            * img_width
        )
        ymin = (
            tf.sparse.to_dense(raw_record["image/object/bbox/ymin"]).numpy()
            * img_height
        )
        ymax = (
            tf.sparse.to_dense(raw_record["image/object/bbox/ymax"]).numpy()
            * img_height
        )

        width = xmax - xmin
        height = ymax - ymin
        bboxes = [
            BBox2D(label, x, y, w, h)
            for label, x, y, w, h in zip(label, xmin, ymin, width, height)
        ]

        return bboxes

    def _load_label_mappings(self):
        label_file = os.path.join(self.root, self.LABEL_FILE)
        with open(label_file, "rb") as f:
            label_map = text_format.Parse(
                f.read(), string_int_label_map_pb2.StringIntLabelMap()
            )

        label_mapppings = {item.id: item.name for item in label_map.item}

        return label_mapppings



import logging

import click
from yacs.config import CfgNode as CN

import datasetinsights.constants as const
from datasetinsights.estimators.base import create_estimator

logger = logging.getLogger(__name__)


@click.command(
    help="Start model evaluation tasks for a pre-trained model.",
    context_settings=const.CONTEXT_SETTINGS,
)
@click.option(
    "-c",
    "--config",
    type=click.STRING,
    required=True,
    help="Path to the config estimator yaml file.",
)
@click.option(
    "-p",
    "--checkpoint-file",
    type=click.STRING,
    required=True,
    help="URI to a checkpoint file.",
)
@click.option(
    "-t",
    "--test-data",
    type=click.Path(exists=True, file_okay=False),
    required=True,
    help="Directory on localhost where test dataset is located.",
)
@click.option(
    "-l",
    "--tb-log-dir",
    type=click.STRING,
    default=const.DEFAULT_TENSORBOARD_LOG_DIR,
    help=(
        "Path to the directory where tensorboard events should be stored. "
        "This Path can be GCS URI (e.g. gs://<bucket>/runs) or full path "
        "to a local directory."
    ),
)
@click.option(
    "-w",
    "--workers",
    type=click.INT,
    default=0,
    help=(
        "Number of multiprocessing workers for loading datasets. "
        "Set this argument to 0 will disable multiprocessing which is "
        "recommended when running inside a docker container."
    ),
)
@click.option(
    "--kfp-metrics-dir",
    type=click.Path(file_okay=False, writable=True),
    default=const.DEFAULT_KFP_METRICS_DIR,
    help="Path to the directory where Kubeflow Metrics files are stored.",
)
@click.option(
    "--kfp-metrics-filename",
    type=click.STRING,
    default=const.DEFAULT_KFP_METRICS_FILENAME,
    help="Kubeflow Metrics filename.",
)
@click.option(
    "--no-cuda",
    is_flag=True,
    default=False,
    help=(
        "Force to disable CUDA. If CUDA is available and this flag is False, "
        "model will be trained using CUDA."
    ),
)
def cli(
    config,
    checkpoint_file,
    test_data,
    tb_log_dir,
    workers,
    kfp_metrics_dir,
    kfp_metrics_filename,
    no_cuda,
):
    ctx = click.get_current_context()
    logger.debug(f"Called evaluate command with parameters: {ctx.params}")
    logger.debug(f"Override estimator config with args: {ctx.args}")
    config = CN.load_cfg(open(config, "r"))
    estimator = create_estimator(
        config=config,
        name=config.estimator,
        checkpoint_file=checkpoint_file,
        tb_log_dir=tb_log_dir,
        workers=workers,
        kfp_metrics_dir=kfp_metrics_dir,
        kfp_metrics_filename=kfp_metrics_filename,
        no_cuda=no_cuda,
    )

    estimator.evaluate(test_data=test_data)



r"""Average Log10 Error metric.

The average relative error can be described as:

.. math::
    (\delta_i):\%\:of\:y_p\:s.t.\:
    max(\frac{y_p}{\hat{y_p}})=\delta<thr\:for\:thr=1.25,1.25^2,1.25^3
"""
import numpy as np

from .base import EvaluationMetric
from .exceptions import NoSampleError


class ThresholdAccuracy(EvaluationMetric):
    """Threshold accuracy metric.

    The metric is defined for grayscale depth images.

    Attributes:
        sum_of_threshold_acc (int): the sum of threshold accuracies for all the
         images in a branch
        num_samples (int): the number of samples in all mini-batches
    """

    def __init__(self, threshold):
        self.threshold = threshold
        self.sum_of_threshold_acc = 0
        self.num_samples = 0

    def reset(self):
        self.sum_of_threshold_acc = 0
        self.num_samples = 0

    def update(self, output):
        y_pred, y = output
        if y.shape != y_pred.shape:
            raise ValueError(
                f"The shapes of real data {y.shape} and predicted data "
                f"{y_pred.shape} should be the same."
            )

        thresh_scores = np.maximum(y / (y_pred + 1e-15), y_pred / (y + 1e-15))
        self.num_samples += y.shape[0]
        self.sum_of_threshold_acc += (
            np.sum(thresh_scores < self.threshold) / y.size * y.shape[0]
        )

    def compute(self):
        if self.num_samples == 0:
            raise NoSampleError(
                "ThresholdAccuracy must have at least one example before it "
                "can be computed."
            )
        return self.sum_of_threshold_acc / self.num_samples



from .bbox import BBox2D
from .checkpoint import EstimatorCheckpoint
from .downloader import create_downloader
from .kfp_output import KubeflowPipelineWriter

__all__ = [
    "BBox2D",
    "EstimatorCheckpoint",
    "KubeflowPipelineWriter",
    "create_downloader",
]



from .records import Records


def prediction_records(gt_bboxes, pred_bboxes, iou_thresh=0.5):
    """ Calculate prediction results per image.

    Args:
        gt_bboxes (List[BBox2D]): a list of ground truth bounding boxes.
        pred_bboxes (List[BBox2D]): a list of predicted bounding boxes.
        iou_thresh (float): iou threshold. Defaults to 0.5.

    Returns:
        Records: a Records class contains match results.

    """
    records = Records(iou_threshold=iou_thresh)
    records.add_records(gt_bboxes, pred_bboxes)
    return records


def precision_recall(gt_bboxes, pred_bboxes, iou_thresh=0.5):
    """ Calculate precision and recall per image.

    Args:
        gt_bboxes (List[BBox2D]): a list of ground truth bounding boxes.
        pred_bboxes (List[BBox2D]): a list of predicted bounding boxes.
        iou_thresh (float): iou threshold. Defaults to 0.5.

    Returns:
        tuple: (precision_per_image, recall_per_image).

    """
    records = prediction_records(gt_bboxes, pred_bboxes, iou_thresh)
    match_results = records.match_results
    tp_count = 0
    for score, res in match_results:
        if res:
            tp_count += 1
    return tp_count / (len(match_results)), tp_count / len(gt_bboxes)



"""Dummy dataset for testing."""

import numpy as np
from PIL import Image

from datasetinsights.datasets import Dataset
from datasetinsights.io.bbox import BBox2D


class DummyDetection2D(Dataset):
    """Dummy dataset creation with 2 images for testing.

    Attributes:
        images: list of images of dataset
        bboxes: target bounding boxes
        label_mappings: class label mappings
        transform: transform the images and bounding boxes

    """

    def __init__(self, image_size=(256, 256), transform=None):
        """initiate dataset class.

        Args:
            image_size : size of images you want generate
            transform : transform the images and bounding boxes
        """
        self.images = [
            Image.fromarray(np.random.random(image_size), "L"),
            Image.fromarray(np.random.random(image_size), "L"),
        ]
        self.bboxes = [
            [
                BBox2D(label=1, x=10, y=20, w=30, h=40),
                BBox2D(label=2, x=50, y=50, w=10, h=10),
            ],
            [
                BBox2D(label=1, x=30, y=40, w=20, h=20),
                BBox2D(label=2, x=20, y=10, w=40, h=10),
            ],
        ]
        self.label_mappings = {"1": "car", "2": "bike"}

        self.transform = transform

    def __len__(self):
        """get dataset length."""
        return len(self.images)

    def __getitem__(self, index):
        """get data point."""
        image, bboxes = self.images[index], self.bboxes[index]

        if self.transform:
            image, bboxes = self.transform(image, bboxes)
        return image, bboxes

    def download(self):
        """pass download."""
        pass



import random

import torchvision.transforms.functional as F


class Compose:
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, image, target):
        for t in self.transforms:
            image, target = t(image, target)
        return image, target


class RandomHorizontalFlip:
    """Flip the image from top to bottom.

    Args:
        flip_prob: the probability to flip the image
    """

    def __init__(self, flip_prob=0.5):
        self.flip_prob = flip_prob

    def __call__(self, sample):
        image, depth = sample
        if random.random() < self.flip_prob:
            image = F.hflip(image)
            depth = F.hflip(depth)

        return (image, depth)


class Resize:
    """Resize the (image, target) to the given sizes.

    Args:
        img_size (tuple or int): Desired output size. If size is a
            sequence like (h, w), output size will be matched to this. If size
            is an int, smaller edge of the image will be matched to this
            number. i.e, if height > width, then image will be rescaled to
            (size * height / width, size)
        target_size (tuple or int): Desired output size. If size is a
            sequence like (h, w), output size will be matched to this. If size
            is an int, smaller edge of the image will be matched to this
            number. i.e, if height > width, then image will be rescaled to
            (size * height / width, size)
    """

    def __init__(self, img_size=-1, target_size=-1):
        self.img_size = img_size
        self.target_size = target_size

    def __call__(self, sample):
        img, target = sample
        if self.img_size != -1:
            img = F.resize(img, self.img_size)
        if self.target_size != -1:
            target = F.resize(target, self.target_size)
        return (img, target)



r"""Reference.

https://github.com/rafaelpadilla/Object-Detection-Metrics#average-precision\
Update algorithm from:
https://github.com/rafaelpadilla/Object-Detection-Metrics/blob/master/lib/Evaluator.py
"""
import collections

import numpy as np

from datasetinsights.io.bbox import group_bbox2d_per_label

from .base import EvaluationMetric
from .records import Records


class AveragePrecision(EvaluationMetric):
    """2D Bounding Box Average Precision metrics.

    This metric would calculate average precision (AP) for each label under
    an iou threshold (default: 0.5). The maximum number of detections
    per image is limited (default: 100).

    Args:
        iou_threshold (float): iou threshold (default: 0.5)
        interpolation (string): AP interoperation method name for AP calculation
        max_detections (int): max detections per image (default: 100)
    """

    TYPE = "metric_per_label"

    def __init__(
        self,
        iou_threshold=0.5,
        interpolation="EveryPointInterpolation",
        max_detections=100,
    ):
        if interpolation == "EveryPointInterpolation":
            self._ap_method = self.every_point_interpolated_ap
        elif interpolation == "NPointInterpolatedAP":
            self._ap_method = self.n_point_interpolated_ap
        else:
            raise ValueError(f"Unknown AP method name: {interpolation}!")

        self._iou_threshold = iou_threshold
        self._max_detections = max_detections
        self._label_records = collections.defaultdict(
            lambda: Records(iou_threshold=self._iou_threshold)
        )
        self._gt_bboxes_count = collections.defaultdict(int)

    def reset(self):
        """Reset AP metrics."""
        self._label_records = collections.defaultdict(
            lambda: Records(iou_threshold=self._iou_threshold)
        )
        self._gt_bboxes_count = collections.defaultdict(int)

    def update(self, mini_batch):
        """Update records per mini batch

        Args:
            mini_batch (list(list)): a list which contains batch_size of
            gt bboxes and pred bboxes pair in each image.
            For example, if batch size = 2, mini_batch looks like:
            [[gt_bboxes1, pred_bboxes1], [gt_bboxes2, pred_bboxes2]]
            where gt_bboxes1, pred_bboxes1 contain gt bboxes and pred bboxes
            in one image
        """
        for bboxes in mini_batch:
            gt_bboxes, pred_bboxes = bboxes

            pred_bboxes = sorted(
                pred_bboxes, key=lambda bbox: bbox.score, reverse=True
            )
            if len(pred_bboxes) > self._max_detections:
                pred_bboxes = pred_bboxes[: self._max_detections]

            bboxes_per_label = group_bbox2d_per_label(pred_bboxes)
            for label, boxes in bboxes_per_label.items():
                self._label_records[label].add_records(gt_bboxes, boxes)

            for gt_bbox in gt_bboxes:
                self._gt_bboxes_count[gt_bbox.label] += 1

    def compute(self):
        """Compute AP for each label.

        Returns:
            dict: a dictionary of AP scores per label.
        """
        average_precision = {}
        _label_records = self._label_records

        for label in self._gt_bboxes_count:
            # if there are no predicted boxes with this label
            if label not in _label_records:
                average_precision[label] = 0
                continue
            match_results = _label_records[label].match_results
            _gt_bboxes_count = self._gt_bboxes_count[label]

            match_results = sorted(match_results, reverse=True)
            true_pos = np.array(list(zip(*match_results))[1]).astype(int)
            false_pos = 1 - true_pos

            acc_tp = np.cumsum(true_pos)
            acc_fp = np.cumsum(false_pos)

            recall = acc_tp / _gt_bboxes_count
            precision = np.divide(acc_tp, (acc_fp + acc_tp))
            ap = self._ap_method(recall, precision)
            # add class result in the dictionary to be returned
            average_precision[label] = ap

        return average_precision

    @staticmethod
    def every_point_interpolated_ap(recall, precision):
        """Calculating the interpolation performed in all points.

        Args:
            recall (list): recall history of the prediction
            precision (list): precision history of the prediction

        Returns:
            float: average precision for all points interpolation
        """
        # TODO: make it readable
        mrec = [0] + list(recall) + [1]
        mpre = [0] + list(precision) + [0]
        for i in range(len(mpre) - 1, 0, -1):
            mpre[i - 1] = max(mpre[i - 1], mpre[i])
        ii = []
        for i in range(len(mrec) - 1):
            if mrec[1:][i] != mrec[0:-1][i]:
                ii.append(i + 1)
        ap = 0
        for i in ii:
            ap = ap + np.sum((mrec[i] - mrec[i - 1]) * mpre[i])

        return ap

    @staticmethod
    def n_point_interpolated_ap(recall, precision, point=11):
        """Calculating the n-point interpolation.

        Args:
            recall (list): recall history of the prediction
            precision (list): precision history of the prediction
            point (int): n, n-point interpolation

        Returns:
            float: average precision for n-point interpolation
        """
        # TODO: make it readable
        mrec = [e for e in recall]
        mpre = [e for e in precision]
        recall_values = np.linspace(0, 1, point)
        recall_values = list(recall_values[::-1])
        rho_interp = []
        recall_valid = []
        # For each recall_values (0, 0.1, 0.2, ... , 1)
        for r in recall_values:
            # Obtain all recall values higher or equal than r
            arg_greater_recalls = np.argwhere(mrec[:] >= r)
            pmax = 0
            # If there are recalls above r
            if arg_greater_recalls.size != 0:
                pmax = max(mpre[arg_greater_recalls.min() :])
            recall_valid.append(r)
            rho_interp.append(pmax)
        ap = sum(rho_interp) / point
        # Generating values for the plot
        rvals = []
        rvals.append(recall_valid[0])
        [rvals.append(e) for e in recall_valid]
        rvals.append(0)
        pvals = []
        pvals.append(0)
        [pvals.append(e) for e in rho_interp]
        pvals.append(0)

        cc = []
        for i in range(len(rvals)):
            p = (rvals[i], pvals[i - 1])
            if p not in cc:
                cc.append(p)
            p = (rvals[i], pvals[i])
            if p not in cc:
                cc.append(p)

        return ap


class AveragePrecisionIOU50(EvaluationMetric):
    """2D Bounding Box Average Precision at IOU = 50%.

    This implementation would calculate AP at IOU = 50% for each label.
    """

    TYPE = "metric_per_label"

    def __init__(self):
        self._ap = AveragePrecision(iou_threshold=0.5)

    def reset(self):
        self._ap.reset()

    def update(self, mini_batch):
        self._ap.update(mini_batch)

    def compute(self):
        return self._ap.compute()


class MeanAveragePrecisionIOU50(EvaluationMetric):
    """2D Bounding Box Mean Average Precision metrics at IOU=50%.

    This implementation would calculate mAP at IOU=50%.

    .. math:: mAP^{IoU=50} = mean_{label}AP^{label, IoU=50}
    """

    TYPE = "scalar"

    def __init__(self):
        self._ap = AveragePrecision(iou_threshold=0.5)

    def reset(self):
        self._ap.reset()

    def update(self, mini_batch):
        self._ap.update(mini_batch)

    def compute(self):
        result = self._ap.compute()
        mean_ap = np.mean(
            [result_per_label for result_per_label in result.values()]
        )
        return mean_ap


class MeanAveragePrecisionAverageOverIOU(EvaluationMetric):
    """2D Bounding Box Mean Average Precision metrics.

    This implementation computes Mean Average Precision (mAP) metric,
    which is implemented as the Average Precision average over all
    labels and IOU = 0.5:0.95:0.05. The max detections per image is
    limited to 100.

    .. math:: mAP^{IoU=0.5:0.95:0.05} = mean_{label,IoU}
    AP^{label, IoU=0.5:0.95:0.05}
    """

    TYPE = "scalar"

    IOU_THRESHOULDS = np.linspace(
        0.5, 0.95, np.round((0.95 - 0.5) / 0.05) + 1, endpoint=True
    )

    def __init__(self):
        self._map_per_iou = [
            AveragePrecision(iou) for iou in self.IOU_THRESHOULDS
        ]

    def reset(self):
        [mean_ap.reset() for mean_ap in self._map_per_iou]

    def update(self, mini_batch):
        for mean_ap in self._map_per_iou:
            mean_ap.update(mini_batch)

    def compute(self):
        """Compute mAP over IOU.
        """
        result = np.mean(
            [
                value
                for ap in self._map_per_iou
                for value in ap.compute().values()
            ]
        )
        return result



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: string_int_label_map.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='string_int_label_map.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=b'\n\x1astring_int_label_map.proto\"U\n\x15StringIntLabelMapItem\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\x05\x12\x14\n\x0c\x64isplay_name\x18\x03 \x01(\t\x12\x0c\n\x04path\x18\x04 \x01(\t\"9\n\x11StringIntLabelMap\x12$\n\x04item\x18\x01 \x03(\x0b\x32\x16.StringIntLabelMapItem'
)




_STRINGINTLABELMAPITEM = _descriptor.Descriptor(
  name='StringIntLabelMapItem',
  full_name='StringIntLabelMapItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='StringIntLabelMapItem.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='id', full_name='StringIntLabelMapItem.id', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='StringIntLabelMapItem.display_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='path', full_name='StringIntLabelMapItem.path', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30,
  serialized_end=115,
)


_STRINGINTLABELMAP = _descriptor.Descriptor(
  name='StringIntLabelMap',
  full_name='StringIntLabelMap',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='item', full_name='StringIntLabelMap.item', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=117,
  serialized_end=174,
)

_STRINGINTLABELMAP.fields_by_name['item'].message_type = _STRINGINTLABELMAPITEM
DESCRIPTOR.message_types_by_name['StringIntLabelMapItem'] = _STRINGINTLABELMAPITEM
DESCRIPTOR.message_types_by_name['StringIntLabelMap'] = _STRINGINTLABELMAP
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

StringIntLabelMapItem = _reflection.GeneratedProtocolMessageType('StringIntLabelMapItem', (_message.Message,), {
  'DESCRIPTOR' : _STRINGINTLABELMAPITEM,
  '__module__' : 'string_int_label_map_pb2'
  # @@protoc_insertion_point(class_scope:StringIntLabelMapItem)
  })
_sym_db.RegisterMessage(StringIntLabelMapItem)

StringIntLabelMap = _reflection.GeneratedProtocolMessageType('StringIntLabelMap', (_message.Message,), {
  'DESCRIPTOR' : _STRINGINTLABELMAP,
  '__module__' : 'string_int_label_map_pb2'
  # @@protoc_insertion_point(class_scope:StringIntLabelMap)
  })
_sym_db.RegisterMessage(StringIntLabelMap)


# @@protoc_insertion_point(module_scope)



r"""Root Mean Square Error metrics.

The root mean square error can be described as:

.. math::
    \sqrt{\frac{1}{n}\sum_{p}^{n}{(y_p-\hat{y_p})}^2}
"""
from math import sqrt

import numpy as np

from .base import EvaluationMetric
from .exceptions import NoSampleError


class RootMeanSquareError(EvaluationMetric):
    """Root Mean Square Error metric.

    The metric is defined for grayscale depth images.

    Attributes:
        sum_of_root_mean_square_error (float): the sum of RMSE
        for all the images in a branch
        num_samples (int): the number of samples in all mini-batches
    """

    def __init__(self):
        self.sum_of_root_mean_square_error = 0
        self.num_samples = 0

    def reset(self):
        self.sum_of_root_mean_square_error = 0
        self.num_samples = 0

    def update(self, output):
        y_pred, y = output
        if y.shape != y_pred.shape:
            raise ValueError(
                f"The shapes of real data {y.shape} and predicted data {y_pred.shape} \
                               should be the same."
            )

        self.num_samples += y.shape[0]
        self.sum_of_root_mean_square_error += (
            np.square(y - y_pred).mean() * y.shape[0]
        )

    def compute(self):
        if self.num_samples == 0:
            raise NoSampleError(
                "RootMeanSquareError must have at least one example before it "
                "can be computed."
            )
        return sqrt(self.sum_of_root_mean_square_error / self.num_samples)
